#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_satisfiability.py.

This file contains functions for converting SAT problems to PUBOs.

"""

# for BOOLEAN_MODELS, we can't just `from qubovert import BOOLEAN_MODELS`
# because it causes circular imports, so instead just import qubovert.
import qubovert as qv


__all__ = "BUFFER", "NOT", "AND", "NAND", "OR", "NOR", "XOR", "XNOR"


def BUFFER(x):
    """BUFFER.

    Return the boolean expression for the buffer of ``x``.

    Parameters
    ----------
    x : hashable object or dict (or subclass of dict).
        The expression to buffer.

    Return
    ------
    P : ``qubovert.PUBO`` object or same type as ``type(x)``.
        If ``x`` is a ``qubovert.QUBO``, ``qubovert.PCBO``,
        ``qubovert.utils.QUBOMatrix``, or ``qubovert.utils.PUBOMatrix`` object,
        then ``type(P) == type(x)``. Otherwise, ``type(P) == type(x)``.

    Example
    -------
    >>> from qubovert.sat import BUFFER
    >>> P = BUFFER('x')
    >>> P
    {('x',): 1}
    >>> P.value({'x': 1})
    1
    >>> P.value({'x': 0})
    0
    >>> type(P)
    qubovert._pubo.PUBO

    >>> P = BUFFER({(0, 1): 1})
    >>> P
    {(0, 1): 1}
    >>> P.value({0: 0, 1: 0})
    0
    >>> P.value({0: 0, 1: 1})
    0
    >>> P.value({0: 1, 1: 0})
    0
    >>> P.value({0: 1, 1: 1})
    1
    >>> type(P)
    qubovert._pubo.PUBO

    >>> from qubovert import boolean_var
    >>> x = boolean_var('x')
    >>> P = BUFFER(x)
    >>> P.value({'x': 1})
    1
    >>> P.value({'x': 0})
    0
    >>> type(P)
    qubovert.PCBO

    """
    if isinstance(x, qv.BOOLEAN_MODELS):
        return x.copy()
    return qv.PUBO(x) if isinstance(x, dict) else qv.PUBO({(x,): 1})


def NOT(x):
    """NOT.

    Return the boolean expression for the NOT of ``x``.

    Parameters
    ----------
    x : hashable object or dict (or subclass of dict).
        The expression to buffer.

    Return
    ------
    P : ``qubovert.PUBO`` object or same type as ``type(x)``.
        The boolean expression for the logic operation.
        If ``x`` is a ``qubovert.QUBO``, ``qubovert.PCBO``,
        ``qubovert.utils.QUBOMatrix``, or ``qubovert.utils.PUBOMatrix`` object,
        then ``type(P) == type(x)``. Otherwise, ``type(P) == type(x)``.

    Example
    -------
    >>> from qubovert.sat import NOT
    >>> P = NOT('x')
    >>> P
    {(): 1, ('x',): -1}
    >>> P.value({'x': 1})
    0
    >>> P.value({'x': 0})
    1
    >>> type(P)
    qubovert._pubo.PUBO

    >>> P = NOT({(0, 1): 1})
    >>> P
    {(): 1, (0, 1): -1}
    >>> P.value({0: 0, 1: 0})
    1
    >>> P.value({0: 0, 1: 1})
    1
    >>> P.value({0: 1, 1: 0})
    1
    >>> P.value({0: 1, 1: 1})
    0

    >>> from qubovert import boolean_var
    >>> x = boolean_var('x')
    >>> P = NOT(x)
    >>> P.value({'x': 1})
    0
    >>> P.value({'x': 0})
    1
    >>> type(P)
    qubovert.PCBO

    """
    return 1 - BUFFER(x)


def AND(*variables):
    """AND.

    Return the boolean expression for the AND of the variables.

    Parameters
    ----------
    *variables : arguments.
        ``variables`` can be of arbitrary length. Each variable can be a
        hashable object, which is the label of the boolean variable, or a dict
        (or subclass of dict) representing a boolean expression.

    Return
    ------
    P : ``qubovert.PUBO`` object or same type as ``type(variables[0])``.
        The boolean expression for the logic operation.
        If ``variables[0]`` is a ``qubovert.QUBO``, ``qubovert.PCBO``,
        ``qubovert.utils.QUBOMatrix``, or ``qubovert.utils.PUBOMatrix`` object,
        then ``type(P) == type(variables[0])``. Otherwise,
        ``type(P) == type(variables[0])``.

    Example
    -------
    >>> from qubovert.sat import AND
    >>> P = AND(0, 1)
    >>> P
    {(0, 1): 1}
    >>> P.value({0: 0, 1: 0})
    0
    >>> P.value({0: 0, 1: 1})
    0
    >>> P.value({0: 1, 1: 0})
    0
    >>> P.value({0: 1, 1: 1})
    1
    >>> type(P)
    qubovert._pubo.PUBO

    >>> P = AND({(0, 1): 1}, 'x')  # and of 0, 1, and 'x'.
    >>> P
    {(0, 1, 'x'): 1}
    >>> type(P)
    qubovert._pubo.PUBO

    >>> from qubovert import boolean_var
    >>> x, y = boolean_var('x'), boolean_var('y')
    >>> P = AND(x, y)
    >>> type(P)
    qubovert.PCBO

    """
    if not variables:
        P = qv.PUBO() + 1
    else:
        P = 1
        for v in variables:
            P *= BUFFER(v)
    return P


def NAND(*variables):
    """NAND.

    Return the boolean expression for the NAND of the variables. Equivalent
    to ``NOT(AND(*variables))``

    Parameters
    ----------
    *variables : arguments.
        ``variables`` can be of arbitrary length. Each variable can be a
        hashable object, which is the label of the boolean variable, or a dict
        (or subclass of dict) representing a boolean expression.

    Return
    ------
    P : ``qubovert.PUBO`` object or same type as ``type(variables[0])``.
        The boolean expression for the logic operation.
        If ``variables[0]`` is a ``qubovert.QUBO``, ``qubovert.PCBO``,
        ``qubovert.utils.QUBOMatrix``, or ``qubovert.utils.PUBOMatrix`` object,
        then ``type(P) == type(variables[0])``. Otherwise,
        ``type(P) == type(variables[0])``.

    Example
    -------
    >>> from qubovert.sat import NAND
    >>> P = NAND(0, 1)
    >>> P
    {(): 1, (0, 1): -1}
    >>> P.value({0: 0, 1: 0})
    1
    >>> P.value({0: 0, 1: 1})
    1
    >>> P.value({0: 1, 1: 0})
    1
    >>> P.value({0: 1, 1: 1})
    0
    >>> type(P)
    qubovert._pubo.PUBO

    >>> P = NAND({(0, 1): 1}, 'x')  # nand of 0, 1, and 'x'.
    >>> P
    {(): 1, (0, 1, 'x'): -1}
    >>> type(P)
    qubovert._pubo.PUBO

    >>> from qubovert import boolean_var
    >>> x, y = boolean_var('x'), boolean_var('y')
    >>> P = NAND(x, y)
    >>> type(P)
    qubovert.PCBO

    """
    return NOT(AND(*variables))


def OR(*variables):
    """OR.

    Return the boolean expression for the OR of the variables.

    Parameters
    ----------
    *variables : arguments.
        ``variables`` can be of arbitrary length. Each variable can be a
        hashable object, which is the label of the boolean variable, or a dict
        (or subclass of dict) representing a boolean expression.

    Return
    ------
    P : ``qubovert.PUBO`` object or same type as ``type(variables[0])``.
        The boolean expression for the logic operation.
        If ``variables[0]`` is a ``qubovert.QUBO``, ``qubovert.PCBO``,
        ``qubovert.utils.QUBOMatrix``, or ``qubovert.utils.PUBOMatrix`` object,
        then ``type(P) == type(variables[0])``. Otherwise,
        ``type(P) == type(variables[0])``.

    Example
    -------
    >>> from qubovert.sat import OR
    >>> P = OR(0, 1)
    >>> P
    {(0,): 1, (0, 1): -1, (1,): 1}
    >>> P.value({0: 0, 1: 0})
    0
    >>> P.value({0: 0, 1: 1})
    1
    >>> P.value({0: 1, 1: 0})
    1
    >>> P.value({0: 1, 1: 1})
    1
    >>> type(P)
    qubovert._pubo.PUBO

    >>> P = OR({(0, 1): 1}, 'x')  # or of 0, 1, and 'x'.
    >>> P
    {(0, 1): 1, (0, 1, 'x'): -1, ('x',): 1}
    >>> type(P)
    qubovert._pubo.PUBO

    >>> from qubovert import boolean_var
    >>> x, y = boolean_var('x'), boolean_var('y')
    >>> P = OR(x, y)
    >>> type(P)
    qubovert.PCBO

    """
    if not variables:
        return qv.PUBO() + 1
    elif len(variables) == 1:
        return BUFFER(variables[0])
    x, v = OR(*variables[:-1]), BUFFER(variables[-1])
    return x + v * (1 - x)


def NOR(*variables):
    """NOR.

    Return the boolean expression for the OR of the variables. Equivalent to
    ``NOT(OR(*variables))``.

    Parameters
    ----------
    *variables : arguments.
        ``variables`` can be of arbitrary length. Each variable can be a
        hashable object, which is the label of the boolean variable, or a dict
        (or subclass of dict) representing a boolean expression.

    Return
    ------
    P : ``qubovert.PUBO`` object or same type as ``type(variables[0])``.
        The boolean expression for the logic operation.
        If ``variables[0]`` is a ``qubovert.QUBO``, ``qubovert.PCBO``,
        ``qubovert.utils.QUBOMatrix``, or ``qubovert.utils.PUBOMatrix`` object,
        then ``type(P) == type(variables[0])``. Otherwise,
        ``type(P) == type(variables[0])``.

    Example
    -------
    >>> from qubovert.sat import NOR
    >>> P = NOR(0, 1)
    >>> P
    {(0,): -1, (0, 1): 1, (1,): -1, (): 1}
    >>> P.value({0: 0, 1: 0})
    1
    >>> P.value({0: 0, 1: 1})
    0
    >>> P.value({0: 1, 1: 0})
    0
    >>> P.value({0: 1, 1: 1})
    0
    >>> type(P)
    qubovert._pubo.PUBO

    >>> P = NOR({(0, 1): 1}, 'x')  # nor of 0, 1, and 'x'.
    >>> P
    {(0, 1): -1, (0, 1, 'x'): 1, ('x',): -1, (): 1}
    >>> type(P)
    qubovert._pubo.PUBO

    >>> from qubovert import boolean_var
    >>> x, y = boolean_var('x'), boolean_var('y')
    >>> P = NOR(x, y)
    >>> type(P)
    qubovert.PCBO

    """
    return NOT(OR(*variables))


def XOR(*variables):
    """XOR.

    Return the boolean expression for the XOR of the variables. XOR(a, b) is 1
    if ``a == b``, otherwise it is 0. ``qubovert`` uses the convention that
    an XOR on > 2 bits is a parity gate. Ie
    ``XOR(0, 1, 2, 3) == XOR(XOR(XOR(0, 1), 2), 3)``.

    Parameters
    ----------
    *variables : arguments.
        ``variables`` can be of arbitrary length. Each variable can be a
        hashable object, which is the label of the boolean variable, or a dict
        (or subclass of dict) representing a boolean expression.

    Return
    ------
    P : ``qubovert.PUBO`` object or same type as ``type(variables[0])``.
        The boolean expression for the logic operation.
        If ``variables[0]`` is a ``qubovert.QUBO``, ``qubovert.PCBO``,
        ``qubovert.utils.QUBOMatrix``, or ``qubovert.utils.PUBOMatrix`` object,
        then ``type(P) == type(variables[0])``. Otherwise,
        ``type(P) == type(variables[0])``.

    Example
    -------
    >>> from qubovert.sat import XOR
    >>> P = XOR(0, 1)
    >>> P
    {(0,): 1, (0, 1): -2, (1,): 1}
    >>> P.value({0: 0, 1: 0})
    0
    >>> P.value({0: 0, 1: 1})
    1
    >>> P.value({0: 1, 1: 0})
    1
    >>> P.value({0: 1, 1: 1})
    0
    >>> type(P)
    qubovert._pubo.PUBO

    >>> P = XOR({(0, 1): 1}, 'x')  # xor of 0, 1, and 'x'.
    >>> P
    {(0, 1): 1, (0, 1, 'x'): -2, ('x',): 1}
    >>> type(P)
    qubovert._pubo.PUBO

    >>> from qubovert import boolean_var
    >>> x, y = boolean_var('x'), boolean_var('y')
    >>> P = XOR(x, y)
    >>> type(P)
    qubovert.PCBO

    The following test will pass.

    >>> for n in range(1, 5):
    >>>     P = XOR(*tuple(range(n)))
    >>>     for i in range(1 << n):
    >>>         sol = decimal_to_boolean(i, n)
    >>>         if sum(sol) % 2 == 1:
    >>>             assert P.value(sol) == 1
    >>>         else:
    >>>             assert not P.value(sol)

    """
    if not variables:
        return qv.PUBO() + 1
    elif len(variables) == 1:
        return BUFFER(variables[0])
    x, v = XOR(*variables[:-1]), BUFFER(variables[-1])
    return (x - v) ** 2


def XNOR(*variables):
    """XNOR.

    Return the boolean expression for the XNOR of the variables. Equivalent to
    ``NOT(XOR(*variables))``.

    Note the following convention.

    XOR(a, b) is 1 if ``a == b``, otherwise it is 0. ``qubovert`` uses the
    convention that an XOR on > 2 bits is a parity gate. Ie
    ``XOR(0, 1, 2, 3) == XOR(XOR(XOR(0, 1), 2), 3)``.

    Parameters
    ----------
    *variables : arguments.
        ``variables`` can be of arbitrary length. Each variable can be a
        hashable object, which is the label of the boolean variable, or a dict
        (or subclass of dict) representing a boolean expression.

    Return
    ------
    P : ``qubovert.PUBO`` object or same type as ``type(variables[0])``.
        The boolean expression for the logic operation.
        If ``variables[0]`` is a ``qubovert.QUBO``, ``qubovert.PCBO``,
        ``qubovert.utils.QUBOMatrix``, or ``qubovert.utils.PUBOMatrix`` object,
        then ``type(P) == type(variables[0])``. Otherwise,
        ``type(P) == type(variables[0])``.

    Example
    -------
    >>> from qubovert.sat import XNOR
    >>> P = XNOR(0, 1)
    >>> P
    {(): 1, (0,): -1, (0, 1): 2, (1,): -1}
    >>> P.value({0: 0, 1: 0})
    1
    >>> P.value({0: 0, 1: 1})
    0
    >>> P.value({0: 1, 1: 0})
    0
    >>> P.value({0: 1, 1: 1})
    1
    >>> type(P)
    qubovert._pubo.PUBO

    >>> P = XNOR({(0, 1): 1}, 'x')  # xnor of 0, 1, and 'x'.
    >>> P
    {(): 1, (0, 1): -1, (0, 1, 'x'): 2, ('x',): -1}
    >>> type(P)
    qubovert._pubo.PUBO

    >>> from qubovert import boolean_var
    >>> x, y = boolean_var('x'), boolean_var('y')
    >>> P = XNOR(x, y)
    >>> type(P)
    qubovert.PCBO

    The following test will pass.

    >>> for n in range(1, 5):
    >>>     P = XNOR(*tuple(range(n)))
    >>>     for i in range(1 << n):
    >>>         sol = decimal_to_boolean(i, n)
    >>>         if sum(sol) % 2 == 1:
    >>>             assert not P.value(sol)
    >>>         else:
    >>>             assert P.value(sol) == 1

    """
    return NOT(XOR(*variables))



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_binary_helpers.py.

This file contains the multiple generic helper functions.

"""

from numpy import log2, ceil


__all__ = 'solution_type', 'num_bits'


def solution_type(solution, default='bool'):
    """solution_type.

    Figure out if the ``solution`` is a solution to a boolean or a spin model.
    If it cannot be determined (ie if ``solution`` is all 1s), then return
    ``default``.

    Parameters
    ----------
    solution : dict or iterable.
        Either a dictionary that maps variable names to values, or an iterable
        with values. If ``solution`` is a bin type, then all the values should
        be either a 0 or 1. If ``solution`` is a spin type, then all the values
        should be either a 1 or -1.
    default : str (optional, defaults to ``'bool'``).
        The default answer to return if the solution type cannot be determined.

    Returns
    -------
    res : str.
        If it is determined that ``solution`` is the solution to a spin model,
        then ``res`` will be ``'spin'``. If it is determined that ``solution``
        is the solution to a boolean model, then `res`` will be ``'bool'``.
        Otherwise, ``res`` will be ``default``.

    Examples
    --------
    >>> solution_type((0, 1, 1, 0))
    'bool'

    >>> solution_type((1, -1, -1, 1))
    'spin'

    In these cases, the default is invoked.

    >>> solution_type((1, 1, 1, 1))
    'bool'
    >>> solution_type((1, 1, 1, 1), default='bool')
    'bool'
    >>> solution_type((1, 1, 1, 1), default=='spin')
    'spin'

    """
    sol = solution.values() if isinstance(solution, dict) else solution
    for v in sol:
        if v == 0:
            return 'bool'
        elif v == -1:
            return 'spin'
    return default


def num_bits(val, log_trick=True):
    """num_bits.

    Find the number of bits needed to represent the value ``val``.

    Parameters
    ----------
    val : numeric.
    log_trick : bool (optional, defaults to True).
        Whether or not to use a log encoding.

    Return
    ------
    num : int.
        The number of bits needed to encode the number ``val``.

    Examples
    --------
    >>> num_bits(7)
    3
    >>> num_bits(8)
    4
    >>> num_bits(7, log_trick=False)
    7
    >>> num_bits(8, log_trick=False)
    8

    """
    if val < 0:
        raise ValueError("``val`` must be >= 0")
    return int(ceil(log2(val + 1) if log_trick else val))



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``utils`` contains many utilities and helpers.

See ``__all__`` for a list of the utilities.

"""

# import order here is important!
from ._warn import *
from ._binary_helpers import *
from ._approximate_extrema import *
from ._hash import *
from ._subgraph import *
from ._normalize import *
from ._solve_bruteforce import *
from ._dict_arithmetic import *
from ._pubomatrix import *
from ._pusomatrix import *
from ._qubomatrix import *
from ._qusomatrix import *
from ._conversions import *
from ._bo_parentclass import *

from ._warn import __all__ as __all_warn__
from ._binary_helpers import __all__ as __all_bh__
from ._approximate_extrema import __all__ as __all_ae__
from ._hash import __all__ as __all_hash__
from ._subgraph import __all__ as __all_subgraph__
from ._normalize import __all__ as __all_normalize__
from ._solve_bruteforce import __all__ as __all_solve_bruteforce__
from ._dict_arithmetic import __all__ as __all_dict_arithmetic__
from ._pubomatrix import __all__ as __all_pubomatrix__
from ._pusomatrix import __all__ as __all_pusomatrix__
from ._qubomatrix import __all__ as __all_qubomatrix__
from ._qusomatrix import __all__ as __all_qusomatrix__
from ._conversions import __all__ as __all_conversions__
from ._bo_parentclass import __all__ as __all_bo__


__all__ = (
    __all_warn__ +
    __all_bh__ +
    __all_ae__ +
    __all_hash__ +
    __all_subgraph__ +
    __all_normalize__ +
    __all_solve_bruteforce__ +
    __all_dict_arithmetic__ +
    __all_pubomatrix__ +
    __all_pusomatrix__ +
    __all_qubomatrix__ +
    __all_qusomatrix__ +
    __all_conversions__ +
    __all_bo__
)


name = "utils"



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_pubo.py.

Contains the PUBO class. See ``help(qubovert.PUBO)``.

"""

from collections import defaultdict
from .utils import BO, PUBOMatrix, QUBOMatrix
from . import QUBO
# in PUBO._reduce_degree, we use PCBO.add_constraint_AND. But PCBO inherits
# from PUBO, so can't say `from . import PCBO` here. Instead, just import
# qubovert
import qubovert as qv


__all__ = 'PUBO',


class PUBO(BO, PUBOMatrix):
    """PUBO.

    Class to manage converting general PUBO problems to and from their
    PUBO, PUSO, QUBO, and QUSO formluations. In general, this class
    deals with unconstrained optimization problems that have arbitrary degree.
    To convert this to a QUBO (see ``to_qubo``) or QUSO (``to_quso``) we have
    to introduce ancilla variables. The ``convert_solution`` method deals with
    converting a solution to the problem with ancilla variables back to the
    solution to the original problem.

    This class deals with PUBOs that have boolean labels that do not range from
    0 to n-1. Note that it is generally
    more efficient to initialize an empty PUBO object and then build the
    PUBO, rather than initialize a PUBO object with an already built dict.

    PUBO inherits some methods and attributes from the ``PUBOMatrix`` class.
    See ``help(qubovert.utils.PUBOMatrix)``.

    PUBO inherits some methods and attributes the ``BO`` class. See
    ``help(qubovert.utils.BO)``.

    Example
    -------
    >>> pubo = PUBO()
    >>> pubo[('a',)] += 5
    >>> pubo[(0, 'a', 1)] -= 2
    >>> pubo -= 1.5
    >>> pubo
    {('a',): 5, ('a', 0, 1): -2, (): -1.5}

    >>> pubo = PUBO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
    >>> pubo
    {('a',): 5, ('a', 0, 1): -2, (): -1.5}
    >>> P = pubo.to_pubo()
    >>> P
    {(0,): 5, (0, 1): -2, (): -1.5}
    >>> pubo.convert_solution({0: 1, 1: 0, 2: 1})
    {'a': 1, 0: 0, 1: 1}

    In the next example, notice that we introduce ancilla variables to
    represent that ```(0, 1)`` term. See the ``to_qubo`` method for more info.

    >>> pubo = PUBO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
    >>> pubo.mapping
    {'a': 0, 0: 1, 1: 2}
    >>> Q = pubo.to_qubo()
    >>> Q
    {(0,): 5, (3,): 9, (0, 1): 3, (0, 3): -6, (1, 3): -6, (2, 3): -2, (): -1.5}
    >>> pubo.convert_solution({0: 1, 1: 0, 2: 1, 2: 0})
    {'a': 1, 0: 0, 1: 1}

    Note
    ----
    Note that keys will end up sorted by their hash. Hashes will not be
    consistent across Python sessions (unless they are integers)! For example,
    both of the following can happen:

    >>> print(PUBO({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (0, 1): -1}

    or

    >>> print(PUBO({('a', 0): 1, (0, 1): -1}))
    {(0, 'a'): 1, (0, 1): -1}

    But the following will never happen:

    >>> print(PUBO({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (1, 0): -1}

    Ie integers will always be correctly sorted.

    Note
    ----
    For efficiency, many internal variables including mappings are computed as
    the problemis being built. This can cause these
    values to be wrong for some specific situations. Calling ``refresh``
    will rebuild the dictionary, resetting all of the values. See
    ``help(PUBO.refresh)``

    Examples
    --------
    >>> from qubovert import PUBO
    >>> P = PUBO()
    >>> P[('a',)] += 1
    >>> P, P.mapping, P.reverse_mapping
    {('a',): 1}, {'a': 0}, {0: 'a'}
    >>> P[('a',)] -= 1
    >>> P, P.mapping, P.reverse_mapping
    {}, {'a': 0}, {0: 'a'}
    >>> P.refresh()
    >>> P, P.mapping, P.reverse_mapping
    {}, {}, {}

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with PUBOs that have boolean labels that do not range
        from 0 to n-1. Note that it is generally more efficient
        to initialize an empty PUBO object and then build the PUBO, rather than
        initialize a PUBO object with an already built dict.

        Parameters
        ----------
        arguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class.

        Examples
        --------
        >>> pubo = PUBO()
        >>> pubo[('a',)] += 5
        >>> pubo[(0, 'a')] -= 2
        >>> pubo -= 1.5
        >>> pubo
        {('a',): 5, ('a', 0): -2, (): -1.5}

        >>> pubo = PUBO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> pubo
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}

        """
        BO.__init__(self, *args, **kwargs)
        PUBOMatrix.__init__(self, *args, **kwargs)

    @staticmethod
    def default_lam(v):
        """default_lam.

        This is the default function used in ``to_qubo``. It returns
        ``1 + abs(v)``. It weights the penalties used to enforce the constraint
        ``xy = z``. See the ``to_qubo`` method.

        Parameters
        ----------
        v : float.

        Return
        ------
        res : float.
            Penalty weight.

        """
        return 1 + abs(v)

    def _reduce_degree(self, D, deg, lam, pairs):
        """_reduce_degree.

        Reduce the degree of the higher order model to a degree ``deg`` model.
        This is a private used only internally and in ``qubovert.PUSO``. See
        ``qubovert.PUBO.to_pubo``, ``qubovert.PUSO.to_puso``,
        ``qubovert.PUBO.to_qubo``, and ``qubovert.PUSO.to_quso`` to see
        an example of this function being used.

        The labels will be integers from 0 to n-1. We introduce ancilla
        variables in order to reduce the degree. The solution to the higher
        order model can be read from the solution to the lower degree model by
        using the ``self.convert_solution`` method.

        Parameters
        ----------
        D : ``qubovert.utils.QUBOMatrix`` or ``qubovert.utils.QUSOMatrix``.
            The dictionary to fill. For reducing PUBOs to QUBOs, ``D`` should
            be a ``qubovert.utils.QUBOMatrix`` object. For reducing PUSOs
            to QUSOs, ``D`` should be a ``qubovert.utils.QUSOMatrix`` object.
            ``D`` should be empty to start.
        deg : int >= 2.
            The degree of the model to reduce to. If``deg`` is None, then
            the model's degree will not be reduced.
        lam : function, numeric, or symbol.
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the model, we add
            penalties to the lower order model in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the higher
            order model may be reduced to a term ``(0, 3): 3`` for the lower
            order model, and then the fact that ``3`` should be the product
            of ``1`` and ``2`` will be enforced with a penalty weight
            ``lam(3)``. If ``lam`` is a number of symbol, then it will be used
            as described above, where ``lam(v) = lam``.
        pairs : set.
            A set of tuples of variable pairs to prioritize pairing together in
            to degree reduction. If a pair in ``pairs`` is found together in
            the PUBO, it will be chosen as a pair to reduce to a single
            ancilla. You should supply this parameter if you have a good idea
            of an efficient way to reduce the degree of the PUBO. If ``pairs``
            is None, then it will be the empty set ``set()``. In other words,
            no variable pairs will be prioritized, and instead variable pairs
            will be chosen to reduce to an ancilla bases solely on frequency
            of occurrance.

        Return
        ------
        None. ``D`` will be updated in place!

        """
        if deg is not None and deg < 2:
            raise ValueError("deg must be >= 2")
        if deg is None:
            deg = self.degree

        if lam is None:
            func_lam = PUBO.default_lam
        elif callable(lam):
            func_lam = lam
        else:
            def func_lam(v): return lam

        # map the pair variables
        pairs = {
            tuple(sorted(self._mapping[i] for i in p))
            if all(i in self._mapping for i in p) else ()
            for p in pairs or {}
        }

        # determine the most common pairs
        pair_frequencies, mapped_self = defaultdict(int), {}
        for k, v in self.items():
            key = tuple(sorted(self._mapping[i] for i in k))
            mapped_self[key] = mapped_self.get(key, 0) + v
            len_key = len(key)
            for i in range(len_key):
                for j in range(i + 1, len_key):
                    pair = key[i], key[j]
                    pair_frequencies[pair] += 1

        # next available label
        ancilla = self.num_binary_variables

        # do the reductions
        reductions = {}
        for key, v in mapped_self.items():
            # find a reduction if len(key) > deg
            while len(key) > deg:
                # find a variable pair in k that has already been reduced.
                in_pairs, previously_used = False, False
                best_pair = None, None
                for i, x in enumerate(key[:-1]):
                    for y in key[i+1:]:
                        pair = x, y
                        if pair in reductions:
                            previously_used = True
                            break
                        # prioritize the pairs that the user picked.
                        elif pair in pairs:
                            in_pairs = True
                            best_pair = None, pair
                            break
                        elif (best_pair[0] is None or
                              pair_frequencies[pair] > best_pair[0]):
                            best_pair = pair_frequencies[pair], pair
                    if previously_used or in_pairs:
                        break

                if previously_used:
                    # z is the ancilla variable for this reduction
                    z = reductions[(x, y)]
                else:
                    # we haven't already reduced the
                    # variable pair (x, y), so take the pair that has the
                    # greatest frequency or that is in pairs.
                    x, y = best_pair[1]
                    z = ancilla
                    reductions[(x, y)] = z
                    ancilla += 1
                    pair_frequencies[(x, z)] += 1
                    pair_frequencies[(y, z)] += 1

                # note we add the constraint even if we've already added
                # it before (if found is True). This is because if we use
                # the reduction multiple times, we need to enforce it
                # multiple times.

                # enforce that z == x y
                D += qv.PCBO().add_constraint_eq_AND(
                    z, x, y, lam=func_lam(v)
                )

                # key is sorted, but it is not necessarily the case that
                # z > all of the other elements in key. So let's efficiently
                # figure out where to put z (without calling sorted)

                old_key, key, z_inserted = key, (), False
                for i in old_key:
                    if i in (x, y):
                        continue
                    elif not z_inserted and z < i:
                        key += (z, i)
                        z_inserted = True
                    else:
                        key += (i,)
                if not z_inserted:
                    key += (z,)

            D[key] += v

    def to_pubo(self, deg=None, lam=None, pairs=None):
        """to_pubo.

        Create and return upper triangular degree ``deg`` PUBO representing the
        problem. The labels will be integers from 0 to n-1.

        Parameters
        ----------
        deg : int >= 0 (optional, defaults to None).
            The degree of the final PUBO. If ``deg`` is None, then the degree
            of the output PUBO will be the same as the degree of ``self``,
            ie see ``self.degree``.
        lam : function (optional, defaults to None).
            Note that if ``deg`` is None or ``deg >= self.degree``, then
            ``lam`` is unneccessary and will not be used.
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the model, we add
            penalties to the lower order model in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the higher
            order model may be reduced to a term ``(0, 3): 3`` for the lower
            order model, and then the fact that ``3`` should be the product
            of ``1`` and ``2`` will be enforced with a penalty weight
            ``lam(3)``.
        pairs : set (optional, defaults to None).
            A set of tuples of variable pairs to prioritize pairing together in
            to degree reduction. If a pair in ``pairs`` is found together in
            the PUBO, it will be chosen as a pair to reduce to a single
            ancilla. You should supply this parameter if you have a good idea
            of an efficient way to reduce the degree of the PUBO. If ``pairs``
            is None, then it will be the empty set ``set()``. In other words,
            no variable pairs will be prioritized, and instead variable pairs
            will be chosen to reduce to an ancilla bases solely on frequency
            of occurrance.

        Return
        ------
        P : qubovert.utils.PUBOMatrix object.
            The upper triangular PUBO matrix, a PUBOMatrix object.
            For most practical purposes, you can use PUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUBOMatrix)``.

        Notes
        -----
        The penalty that we use to enforce the constraints that the ancilla
        variable ``z`` is equal to the product of the two variables that it is
        replacing, ``xy``, is:

            ``0`` if ``z == xy``,
            ``3*lam(v)`` if ``x == y == 0 and z == 1``, and
            ``lam(v)`` else.

        See https://arxiv.org/pdf/1307.8041.pdf equation 6.

        """
        P = PUBOMatrix()
        self._reduce_degree(P, deg, lam, pairs)
        return P

    def to_qubo(self, lam=None, pairs=None):
        """to_qubo.

        Create and return upper triangular QUBO representing the problem.
        The labels will be integers from 0 to n-1. We introduce ancilla
        variables in order to reduce the degree of the PUBO to a QUBO. The
        solution to the PUBO can be read from the solution to the QUBO by
        using the ``convert_solution`` method.

        Parameters
        ----------
        lam : function (optional, defaults to None).
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the PUBO to a
            QUBO, we add penalties to the QUBO in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the PUBO
            may be reduced to a term ``(0, 3): 3`` for the QUBO, and then the
            fact that ``3`` should be the product of ``1`` and ``2`` will be
            enforced with a penalty weight ``lam(3)``.
        pairs : set (optional, defaults to None).
            A set of tuples of variable pairs to prioritize pairing together in
            to degree reduction. If a pair in ``pairs`` is found together in
            the PUBO, it will be chosen as a pair to reduce to a single
            ancilla. You should supply this parameter if you have a good idea
            of an efficient way to reduce the degree of the PUBO. If ``pairs``
            is None, then it will be the empty set ``set()``. In other words,
            no variable pairs will be prioritized, and instead variable pairs
            will be chosen to reduce to an ancilla bases solely on frequency
            of occurrance.

        Return
        ------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, a QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUBOMatrix)``.

        """
        Q = QUBOMatrix()
        self._reduce_degree(Q, 2, lam, pairs)
        return Q

    def convert_solution(self, solution, spin=False):
        """convert_solution.

        Convert the solution to the integer labeled PUBO to the solution to
        the originally labeled PUBO.

        Parameters
        ----------
        solution : iterable or dict.
            The PUBO, PUSO, QUBO, or QUSO solution output. The PUBO solution
            output is either a list or tuple where indices specify the label of
            the variable and the element specifies whether it's 0 or 1 for PUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value. The QUBO/QUSO solution output
            includes the assignment for the ancilla variables used to reduce
            the degree of the PUBO.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        res : dict.
            Maps boolean variable labels to their PUBO solutions values {0, 1}.

        Example
        -------
        >>> pubo = PUBO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> pubo
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}
        >>> P = pubo.to_pubo()
        >>> P
        {(0,): 5, (0, 1): -2, (): -1.5}
        >>> pubo.convert_solution({0: 1, 1: 0, 2: 1})
        {'a': 1, 0: 0, 1: 1}

        In the next example, notice that we introduce ancilla variables to
        represent that ```(0, 1)`` term. See the ``to_qubo`` method for more
        info.

        >>> pubo = PUBO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> pubo.mapping
        {'a': 0, 0: 1, 1: 2}
        >>> Q = pubo.to_qubo(3)
        >>> Q
        {(0,): 5, (0, 3):-2, ():-1.5, (1, 2): 3, (3,): 3, (1, 3): 3, (2, 3): 3}
        >>> pubo.convert_solution({0: 1, 1: 0, 2: 1, 2: 0})
        {'a': 1, 0: 0, 1: 1}

        Notes
        -----
        We take ignore the ancilla variable assignments when we convert the
        solution. For example if the conversion from PUBO to QUBO introduced
        an ancilla varable ``z = xy`` where ``x`` and ``y`` are variables of
        the PUBO, then ``solution`` must have values for ``x``, ``y``, and
        ``z``. If the QUBO solver found that ``x = 1``, ``y = 0``, and
        ``z = 1``, then the constraint that ``z = xy`` is not satisfied (one
        possible cause for this is if the ``lam`` argument in ``to_qubo`` is
        too small). ``convert_solution`` will return that ``x = 1`` and
        ``y = 0`` and ignore the value of ``z``.

        """
        # this works for converting a solution to the pubo, qubo, puso, or
        # quso formulations, since in the to_qubo function all ancilla
        # variables are labeled with integers >= self.num_binary_variables.
        return QUBO.convert_solution(self, solution, spin)

    @staticmethod
    def _check_key_valid(key):
        """_check_key_valid.

        Internal method to check if an input key to the dictionary is valid.
        Checks to see if ``key`` is a tuple.

        Parameters
        ----------
        key : anything, but must be a tuple to be valid.

        Returns
        -------
        None.

        Raises
        ------
        KeyError if the key is invalid.

        """
        # override PUBOMatrix._check_key_valid to allow for noninteger keys.
        if not isinstance(key, tuple):
            raise KeyError(
                "Key formatted incorrectly, must be tuple")



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``qubovert`` is a module for converting problems into QUBO/QUSO form.

QUBO stands for Quadratic Unconstrained Boolean Optimization. QUBO problems
have a one-to-one mapping to classical QUSO problems, and most optimization
problems are formatted in QUBO form when the solver is a quantum computer.
See ``qubovert.__all__`` for useful functionality, ``qubovert.problems__all__``
for problems defined, and ``qubovert.utils.__all__`` for some utility
functions, ``qubovert.sat.__all__`` for the satisfiability library, and
``qubovert.sim.__all__`` for the simulation and annealing library.

"""

from ._version import *

from . import utils

from ._qubo import *
from ._quso import *
from ._pubo import *
from ._puso import *
from ._pcbo import *
from ._pcso import *

from ._qubo import __all__ as __all_qubo__
from ._quso import __all__ as __all_quso__
from ._pubo import __all__ as __all_pubo__
from ._puso import __all__ as __all_puso__
from ._pcbo import __all__ as __all_pcbo__
from ._pcso import __all__ as __all_pcso__

__all__ = (
    __all_qubo__ +
    __all_quso__ +
    __all_pubo__ +
    __all_puso__ +
    __all_pcbo__ +
    __all_pcso__
)

from . import sat
from . import sim
from . import problems

BOOLEAN_MODELS = QUBO, PUBO, PCBO, utils.QUBOMatrix, utils.PUBOMatrix
SPIN_MODELS = QUSO, PUSO, PCSO, utils.QUSOMatrix, utils.PUSOMatrix


name = "qubovert"



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""warn.py.

This file contains warning functionality to standardize qubovert's warnings.

"""

import warnings


__all__ = 'QUBOVertWarning',


class QUBOVertWarning(UserWarning):
    """QUBOVertWarning.

    Warning type to standardize qubovert's warnings. Warn with
    ``QUBOVertWarning.warn("message")``.

    """

    @classmethod
    def warn(cls, message):
        r"""warn.

        Parameters
        ----------
        message : str.
            Message to warn with.

        """
        warnings.warn(message, cls, 3)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``sim`` contains ``qubovert``'s simulation and annealing functionality.

See ``__all__`` for a list of uses.

"""

# import order here is important!
from ._puso_simulation import *
from ._pubo_simulation import *
from ._quso_simulation import *
from ._qubo_simulation import *
from ._anneal_results import *
from ._anneal import *

from ._puso_simulation import __all__ as __all_pusosim__
from ._pubo_simulation import __all__ as __all_pubosim__
from ._quso_simulation import __all__ as __all_qusosim__
from ._qubo_simulation import __all__ as __all_qubosim__
from ._anneal_results import __all__ as __all_results__
from ._anneal import __all__ as __all_anneal__


__all__ = (
    __all_pusosim__ +
    __all_pubosim__ +
    __all_qusosim__ +
    __all_qubosim__ +
    __all_results__ +
    __all_anneal__
)


name = "sim"



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_conversions.py.

This file contains methods to convert to and from QUBO/QUSO/PUBO/PUSO
problems, as well as various misc converters

"""

from . import QUBOMatrix, QUSOMatrix, PUBOMatrix, PUSOMatrix
# for QUBO, QUSO, PUBO, PUSO, can't import directly because it will cause
# circular imports, so instead just import qubovert.
import qubovert as qv


__all__ = (
    'qubo_to_quso', 'quso_to_qubo', 'pubo_to_puso', 'puso_to_pubo',
    'Conversions'
)


def qubo_to_quso(Q):
    """qubo_to_quso.

    Convert the specified QUBO problem into an QUSO problem. Note that
    QUBO {0, 1} values go to QUSO {1, -1} values in that order!

    Parameters
    ----------
    Q : dictionary, qubovert.QUBO, or qubovert.utils.QUBOMatrix object.
        Maps tuples of boolean variables indices to the Q value. See
        ``help(qubovert.QUBO)`` and ``help(qubovert.utils.QUBOMatrix)`` for
        info on formatting.

    Returns
    -------
    L : qubovert.QUSO or qubovert.utils.QUSOMatrix object.
        tuple of spin labels map to QUSO values. If ``Q`` is a
        ``qubovert.utils.QUBOMatrix`` object, then ``L`` will be a
        ``qubovert.utils.QUSOMatrix``, otherwise ``L`` will be a
        ``qubovert.QUSO`` object.

    Example
    -------
    >>> Q = {(0,): 1, (0, 1): -1, (1,): 3}
    >>> L = qubo_to_quso(Q)
    >>> isinstance(L, qubovert.utils.QUSOMatrix)
    True

    >>> Q = {('a',): 1, ('a', 'b'): -1, ('b',): 3}
    >>> L = qubo_to_quso(Q)
    >>> isinstance(L, qubovert.QUSO)
    True

    """
    # could just use QUSOMatrix(pubo_to_puso(Q)), but then we spend a lot of
    # time converting from a PUSOMatrix to QUSOMatrix, so instead we
    # explictly write out the conversion.

    # not isinstance! because isinstance(QUBO, QUBOMatrix) is True
    if type(Q) in (QUBOMatrix, qv.QUBO):
        # key will already be squashed
        def squash_key(k): return k
    else:
        squash_key = qv.QUBO.squash_key

    L = QUSOMatrix() if type(Q) == QUBOMatrix else qv.QUBO()

    for kp, v in Q.items():
        k = squash_key(kp)
        if not k:
            L[k] += v
        elif len(k) == 1:
            L[k] -= v / 2
            L[()] += v / 2
        else:
            # len(k) must be 2 because of squash_key
            i, j = k
            L[k] += v / 4
            L[(i,)] -= v / 4
            L[(j,)] -= v / 4
            L[()] += v / 4

    return L


def quso_to_qubo(L):
    """quso_to_qubo.

    Convert the specified QUSO problem into an upper triangular QUBO problem.
    Note that QUSO {1, -1} values go to QUBO {0, 1} values in that order!

    Parameters
    ----------
    L : dictionary, qubovert.QUSO, or qubovert.utils.QUSOMatrix object.
        Tuple of spin labels map to QUSO values. See
        ``help(qubovert.QUSO)`` and ``help(qubovert.utils.QUSOMatrix)`` for
        info on formatting.


    Returns
    -------
    Q : qubovert.QUBO or qubovert.utils.QUBOMatrix object.
        If ``L`` is a ``qubovert.utils.QUSOMatrix`` object, then ``Q`` will be
        a ``qubovert.utils.QUBOMatrix``, otherwise ``Q`` will be a
        ``qubovert.QUBO`` object. See ``help(qubovert.QUBO)`` and
        ``help(qubovert.utils.QUBOMatrix)`` for info on formatting.

    Example
    -------
    >>> L = {(0,): 1, (1,): -1, (0, 1): -1}
    >>> Q = quso_to_qubo(L)
    >>> isinstance(Q, qubovert.utils.QUBOMatrix)
    True

    >>> L = {('a',): 1, ('b',): -1, (0, 1): -1}
    >>> Q = quso_to_qubo(L)
    >>> isinstance(Q, qubovert.QUBO)
    True

    """
    # could just use QUBOMatrix(puso_to_pubo(L)), but then we spend a lot of
    # time converting from a PUBOMatrix to QUBOMatrix, so instead we explictly
    # write out the conversion.

    # not isinstance! because isinstance(QUSO, QUSOMatrix) is True
    if type(L) in (QUSOMatrix, qv.QUSO):
        # key will already be squashed
        def squash_key(k): return k
    else:
        squash_key = qv.QUSO.squash_key

    Q = QUBOMatrix() if type(L) == QUSOMatrix else qv.QUBO()

    for kp, v in L.items():
        k = squash_key(kp)
        if not k:
            Q[k] += v
        elif len(k) == 1:
            Q[k] -= 2 * v
            Q[()] += v
        else:
            # len(k) must be 2 because of squash_key
            i, j = k
            Q[k] += 4 * v
            Q[(i,)] -= 2 * v
            Q[(j,)] -= 2 * v
            Q[()] += v

    return Q


def pubo_to_puso(P):
    """pubo_to_puso.

    Convert the specified PUBO problem into an PUSO problem. Note that
    PUBO {0, 1} values go to PUSO {1, -1} values in that order!

    Parameters
    ----------
    P : dictionary, qubovert.PUBO, or qubovert.utils.PUBOMatrix object.
        Maps tuples of boolean variables indices to the P value. See
        ``help(qubovert.PUBO)`` and ``help(qubovert.utils.PUBOMatrix)`` for
        info on formatting.

    Returns
    -------
    H : qubovert.utils.PUSOMatrix object.
        tuple of spin labels map to PUSO values. If ``P`` is a
        ``qubovert.utils.PUBOMatrix`` object, then ``H`` will be a
        ``qubovert.utils.PUSOMatrix``, otherwise ``H`` will be a
        ``qubovert.PUSO`` object..

    Example
    -------
    >>> P = {(0,): 1, (0, 1): -1, (1,): 3}
    >>> H = pubo_to_puso(P)
    >>> isinstance(H, qubovert.utils.PUSOMatrix)
    True

    >>> P = {(0,): 1, (0, 1): -1, (1,): 3}
    >>> H = pubo_to_puso(P)
    >>> isinstance(H, qubovert.QUSO)
    True

    """
    def generate_new_key_value(k):
        """generate_new_key_value.

        Recursively generate the PUBO key, value pairs for converting the
        product ``x[k[0]] * ... * x[k[-1]]``, where each ``x`` is a boolean
        variable in {0, 1}, to the product
        ``(1-z[k[0]])/2 * ... * (1-z[k[1]])/2``., where each ``z`` is a spin
        in {1, -1}.

        Parameters
        ----------
        k : tuple.
            Each element of the tuple corresponds to a boolean label.

        Yields
        ------
        res : tuple (key, value)
            key : tuple.
                Each element of the tuple corresponds to a spin label.
            value : float.
                The value to multiply the value corresponding with ``k`` by.

        """
        if not k:
            yield k, 1
        else:
            for key, value in generate_new_key_value(k[1:]):
                yield (k[0],) + key, -value / 2
                yield key, value / 2

    # not isinstance! because isinstance(PUBO, PUBOMatrix) is True
    H = PUSOMatrix() if type(P) == PUBOMatrix else qv.PUSO()

    for k, v in P.items():
        for key, value in generate_new_key_value(k):
            H[key] += value * v

    return H


def puso_to_pubo(H):
    """puso_to_pubo.

    Convert the specified PUSO problem into an upper triangular PUBO problem.
    Note that PUSO {1, -1} values go to PUBO {0, 1} values in that order!

    Parameters
    ----------
    H : dictionary or qubovert.utils.PUSOMatrix object.
        Tuple of spin labels map to PUSO values. See
        ``help(qubovert.PUSO)`` and ``help(qubovert.utils.PUSOMatrix)`` for
        info on formatting.

    Returns
    -------
    P : qubovert.utils.PUBOMatrix object.
        If ``H`` is a ``qubovert.utils.PUSOMatrix`` object, then ``P`` will
        be a ``qubovert.utils.PUBOMatrix``, otherwise ``P`` will be a
        ``qubovert.PUBO`` object. See ``help(qubovert.PUBO)`` and
        ``help(qubovert.utils.PUBOMatrix)`` for info on formatting.

    Example
    -------
    >>> H = {(0,): 1, (1,): -1, (0, 1): -1}
    >>> P = puso_to_pubo(H)
    >>> isinstance(P, qubovert.utils.PUBOMatrix)
    True

    >>> H = {('0',): 1, ('1',): -1, (0, 1): -1}
    >>> P = puso_to_pubo(H)
    >>> isinstance(P, qubovert.PUBO)
    True

    """
    def generate_new_key_value(k):
        """generate_new_key_value.

        Recursively generate the PUBO key, value pairs for converting the
        product ``z[k[0]] * ... * z[k[-1]]``, where each ``z`` is a spin in
        {1, -1}, to the product ``(1-2*x[k[0]]) * ... * (1-2*x[k[1]])``, where
        each ``x`` is a boolean variables in {0, 1}.

        Parameters
        ----------
        k : tuple.
            Each element of the tuple corresponds to a spin label.

        Yields
        ------
        res : tuple (key, value)
            key : tuple.
                Each element of the tuple corresponds to a binary label.
            value : float.
                The value to multiply the value corresponding with ``k`` by.

        """
        if not k:
            yield k, 1
        else:
            for key, value in generate_new_key_value(k[1:]):
                yield (k[0],) + key, -2 * value
                yield key, value

    # not isinstance! because isinstance(PUSO, PUSOMatrix) is True
    P = PUBOMatrix() if type(H) == PUSOMatrix else qv.PUBO()

    for k, v in H.items():
        for key, value in generate_new_key_value(k):
            P[key] += value * v

    return P


class Conversions:
    """Conversions.

    This is a parent class that defines the functions ``to_qubo``,
    ``to_quso``, ``to_pubo``, and ``to_puso``. Any subclass that inherits
    from ``Conversions`` `must` supply at least one of ``to_qubo`` or
    ``to_quso``. And at least one of ``to_pubo`` or ``to_puso``.


    """

    def to_qubo(self, *args, **kwargs):
        """to_qubo.

        Create and return upper triangular QUBO representing the problem.
        Should be implemented in child classes. If this method is not
        implemented in the child class, then it simply calls ``to_quso`` and
        converts the quso formulation to a QUBO formulation.

        Parameters
        ----------
        arguments : Defined in the child class.
            They should be parameters that define lagrange multipliers or
            factors in the QUBO.

        Returns
        -------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, a QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUBOMatrix)``.

        Raises
        ------
        ``RecursionError`` if neither ``to_qubo`` nor ``to_quso`` are defined
        in the subclass.

        """
        return quso_to_qubo(self.to_quso(*args, **kwargs))

    def to_quso(self, *args, **kwargs):
        """to_quso.

        Create and return QUSO model representing the problem.
        Should be implemented in child classes. If this method is not
        implemented in the child class, then it simply calls ``to_qubo`` and
        converts the QUBO formulation to an QUSO formulation.

        Parameters
        ----------
        arguments : Defined in the child class.
            They should be parameters that define lagrange multipliers or
            factors in the QUSO model.

        Return
        ------
        L : qubovert.utils.QUSOMatrix object.
            The upper triangular coupling matrix, where two element tuples
            represent couplings and one element tuples represent fields.
            For most practical purposes, you can use IsingCoupling in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUSOMatrix)``.

        Raises
        ------
        ``RecursionError`` if neither ``to_qubo`` nor ``to_quso`` are defined
        in the subclass.

        """
        return qubo_to_quso(self.to_qubo(*args, **kwargs))

    def to_pubo(self, *args, **kwargs):
        """to_pubo.

        Create and return upper triangular PUBO representing the problem.
        Should be implemented in child classes. If this method is not
        implemented in the child class, then it simply calls ``to_puso`` or
        ``to_qubo`` and converts the puso or QUBO formulations to a
        PUBO formulation.

        Parameters
        ----------
        arguments : Defined in the child class.
            They should be parameters that define lagrange multipliers or
            factors in the QUBO.

        Returns
        -------
        P : qubovert.utils.PUBOMatrix object.
            The upper triangular PUBO matrix, a PUBOMatrix object.
            For most practical purposes, you can use PUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUBOMatrix)``.

        Raises
        ------
        ``RecursionError`` if neither ``to_pubo`` nor ``to_puso`` are defined
        in the subclass.

        """
        return puso_to_pubo(self.to_puso(*args, **kwargs))

    def to_puso(self, *args, **kwargs):
        """to_puso.

        Create and return PUSO model representing the problem.
        Should be implemented in child classes. If this method is not
        implemented in the child class, then it simply calls ``to_pubo`` or
        ``to_quso`` and converts to a PUSO formulation.

        Parameters
        ----------
        arguments : Defined in the child class.
            They should be parameters that define lagrange multipliers or
            factors in the QUSO model.

        Return
        ------
        H : qubovert.utils.PUSOMatrix object.
            For most practical purposes, you can use PUSOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUSOMatrix)``.

        Raises
        ------
        ``RecursionError`` if neither ``to_pubo`` nor ``to_puso`` are defined
        in the subclass.

        """
        return pubo_to_puso(self.to_pubo(*args, **kwargs))



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_alternating_sectors_chain.py.

Contains the AlternatingSectorsChain class. See
``help(qubovert.problems.AlternatingSectorsChain)``.
"""

from qubovert.utils import QUSOMatrix, boolean_to_spin, solution_type
from qubovert.problems import Problem


__all__ = 'AlternatingSectorsChain',


class AlternatingSectorsChain(Problem):
    """AlternatingSectorsChain.

    Class to manage converting Alternating Sectors Chain to and from its QUBO
    and QUSO formluations.

    The Alternating Sectors Chain problem has a solution for which
    all the boolean variables or spins are equal. It is a trivial problem,
    but useful for benchmarking some solvers or solving techniques.

    AlternatingSectorsChain inherits some methods and attributes from the
    Problem class. See ``help(qubovert.problems.Problem)``.

    Example
    -------
    >>> from qubovert.problems import AlternatingSectorsChain
    >>> from any_module import qubo_solver
    >>> # or you can use my bruteforce solver...
    >>> # from qubovert.utils import solve_qubo_bruteforce as qubo_solver
    >>> problem = AlternatingSectorsChain(10)
    >>> Q  = problem.to_qubo()
    >>> obj, sol = qubo_solver(Q)
    >>> solution = problem.convert_solution(sol)

    >>> print(solution)
    (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    # or (1, 1, 1, 1, 1, 1, 1, 1, 1, 1)

    >>> print(problem.is_solution_valid(solution))
    True  # since they are all the same.

    """

    def __init__(self, num_binary_variables,
                 chain_length=3, min_strength=1, max_strength=10):
        """__init__.

        The Alternating Sectors Chain problem has a solution for which
        all the boolean variables or spins are equal. It is a trivial problem,
        but useful for benchmarking some solvers or solving techniques.

        Parameters
        ----------
        num_binary_variables : int > 0.
            Number of variables in the problem.
        chain_length : int > 1 (optional, defaults to 3).
            The length of a chain of couples.
        min_strength : int > 0 (optional, defaults to 1).
            The strength of the couples for the minimium chain.
        max_strength : int > 0 (optional, defaults to 10).
            The strength of the couples for the maximum chain.

        Examples
        --------
        >>> args = n, l, min_s, max_s = 6, 3, 1, 5
        >>> problem = AlternatingSectorsChain(*args)
        >>> h, J, offset = problem.to_quso(pbc=True)
        >>> h
        {}
        >>> offset
        0
        >>> J
        {(0, 1): 5, (1, 2): 5, (2, 3): 5, (3, 4): 1, (4, 5): 1, (5, 0): 1}

        """
        self._N, self._chain_length = num_binary_variables, chain_length
        self._min_strength, self._max_strength = -min_strength, -max_strength
        if min_strength < 0 or max_strength < 0:
            raise ValueError("Coupling strengths must be positive")
        elif num_binary_variables < 1:
            raise ValueError("Must have at least 1 binary variable")
        elif chain_length < 2:
            raise ValueError("Chain length must be at least 2")

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        The number of binary variables that the QUBO and QUSO use.

        Return
        ------
        num : integer.
            The number of variables in the QUBO/QUSO formulation.

        """
        return self._N

    def to_quso(self, pbc=False):
        r"""to_quso.

        Create and return the alternating Sectors chain problem in QUSO form
        The J coupling matrix for the QUSO will be returned as an
        uppertriangular dictionary. Thus, the problem becomes minimizing
        :math:`\sum_{i <= j} z_i z_j J_{ij} + \sum_{i} z_i h_i + offset`.

        Parameters
        ----------
        pbc: bool (optional, defaults to False).
            Whether or not to use periodic boundary conditions.

        Return
        ------
        L : qubovert.utils.QUSOMatrix object.
            For most practical purposes, you can use QUSOMatrix in the
            same way as an ordinary dictionary. For more information, see
            ``help(qubovert.utils.QUSOMatrix)``.

        Example
        -------
        >>> args = n, l, min_s, max_s = 6, 3, 1, 5
        >>> problem = AlternatingSectorsChain(*args)
        >>> L = problem.to_quso(pbc=True)
        >>> L
        {(0, 1): 5, (1, 2): 5, (2, 3): 5, (3, 4): 1, (4, 5): 1, (5, 0): 1}

        >>> L = problem.to_quso(pbc=False)
        >>> L
        {(0, 1): 5, (1, 2): 5, (2, 3): 5, (3, 4): 1, (4, 5): 1}

        """
        L = QUSOMatrix()

        for q in range(self._N-1):
            L[(q, q+1)] = (
                self._min_strength
                if (q // self._chain_length) % 2
                else self._max_strength
            )

        if pbc:
            L[(self._N-1, 0)] = (
                self._min_strength
                if ((self._N-1) // self._chain_length) % 2
                else self._max_strength
            )

        return L

    def convert_solution(self, solution, spin=False):
        """convert_solution.

        Convert the solution to the QUBO or QUSO to the solution to the
        Alternating Sectors Chain problem.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        res : tuple.
            Value of each spin, -1 or 1.

        Examples
        --------
        >>> problem = AlternatingSectorsChain(5)
        >>> problem.convert_solution([0, 0, 1, 0, 1])
        (1, 1, -1, 1, -1)
        >>> problem.convert_solution([-1, -1, 1, -1, 1])
        (-1, -1, 1, -1, 1)

        """
        if isinstance(solution, dict):
            solution = tuple(v for _, v in sorted(solution.items()))
        sol_type = solution_type(solution, 'spin' if spin else 'bool')
        if sol_type == 'bool':
            return boolean_to_spin(solution)
        return solution

    def is_solution_valid(self, solution, spin=False):
        """is_solution_valid.

        Returns whether or not the proposed solution is the correct solution,
        ie that all the variables are equal.

        Parameters
        ----------
        solution : iterable or dict.
            solution can be the output of
            AlternatingSectorsChain.convert_solution,
            or the  QUBO or QUSO solver output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        valid : boolean.
            True if the proposed solution is valid, else False.

        """
        if isinstance(solution, dict):
            solution = self.convert_solution(solution, spin)

        return all(
            x == 1 for x in solution
        ) or all(
            x != 1 for x in solution
        )



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``problems`` contains many problems converted to QUBO/QUSO form.

This module contains many all the problems that we convert to QUBO/QUSO form.
We organize all the problems into categories, but we import them globally for
user use. See the ``__all__`` value for all the problems imported. Each of the
problems inherits from the ``Problem`` parent class, see
``qubovert.problems.Problem``.

"""

from ._problem_parentclass import *  # don't add to __all__!

from .np import *
from .benchmarking import *

from .np import __all__ as __all_np__
from .benchmarking import __all__ as __all_benchmarking__

__all__ = (
    __all_np__ + __all_benchmarking__
)

name = "problems"



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_qusomatrix.py.

This file contains the QUSOMatrix object.

"""

from . import PUSOMatrix


__all__ = 'QUSOMatrix',


class QUSOMatrix(PUSOMatrix):
    """QUSOMatrix.

    ``QUSOMatrix`` inherits some methods from ``PUSOMatrix``, see
    ``help(qubovert.utils.PUSOMatrix)``.

    A class to handle QUSO matrices. It is the same thing as a dictionary
    with some methods modified. Note that each key must be a tuple integers
    >= 0.

    One method is that values will always default to 0. Consider the following
    example:

    >>> d = QUSOMatrix()
    >>> print(d[(0,)]) # will print 0
    >>> d[(0,)] += 1
    >>> print(d) # will print {(0,): 1}

    One method of QUSOMatrix is that it will always keep the QUSO
    upper triangular! Consider the following example:

    >>> d = QUSOMatrix()
    >>> d[(1, 0)] += 2
    >>> print(d)
    >>> # will print {(0, 1): 2}

    One method is that if we set an item to 0, it will be removed. Consider
    the following example:

    >>> d = QUSOMatrix()
    >>> d[(0, 1)] += 1
    >>> d[(0, 1)] -= 1
    >>> print(d) # will print {}

    One method is that if we initialize QUSOMatrix with a previous dictionary
    it will be reinitialized to ensure that the QUSOMatrix is upper
    triangular and contains no zero values. Consider the following example:

    >>> d = QUSOMatrix({(0,): 1, (1, 0): 2, (2, 0): 0})
    >>> print(d) # will print {(0,): 1, (0, 1): 2}

    We also change the update method so that it follows all the conventions.

    >>> d = Isingatrix({(0,): 1, (0, 1): 2})
    >>> d.update({(0,): 0, (1, 0): 1, (1,): -1})
    >>> print(d)  # will print {(0, 1): 1, (1,): -1}

    We also include arithmetic, addition, subtraction, scalar division,
    multiplication, and all those in place. For example,

    >>> d = QUSOMatrix((0,)=1, (0, 1)=-2)
    >>> g = d + {(0,): -1}
    >>> print(g) # will print {(0, 1): -2}
    >>> g *= 4
    >>> print(g) # will print {(0, 1): -8}
    >>> g -= {(0, 1): -8}
    >>> print(g) # will print {}

    >>> d = QUSOMatrix({(0, 0): 1, (0, 1): -1})
    >>> print(d)
    {(): 1, (0, 1): -1}
    >>> g = {(0,): -1, (1,): 1}
    >>> d *= g
    >>> print(d)
    {(0,): -2, (1,): 2}

    >>> d = QUSOMatrix({(0, 0): 1, (0, 1): -1})
    >>> print(d ** 2 == d * d)
    True

    Adding or subtracting constants will update the () element of the
    dict.

    >>> d = QUSOMatrix()
    >>> d += 5
    >>> print(d)
    {(): 5}

    Finally, if you try to access a key out of order, it will sort the key. Be
    careful with this, it can cause unexpected behavior if you don't know it.
    For example,

    >>> d = QUSOMatrix()
    >>> d[(0, 1)] += 2
    >>> print(d[(1, 0)])  # will print 2

    """

    @staticmethod
    def _check_key_valid(key):
        """_check_key_valid.

        Internal method to check if an input key to the dictionary is valid.
        Checks to see if ``key`` is a tuple of non negative integers with <=
        2 unique integers after squashing.

        Parameters
        ----------
        key : anything, but must be a tuple to be valid.

        Returns
        -------
        k : tuple.
            The squashed key.

        Raises
        ------
        KeyError if the key is invalid.

        """
        k = PUSOMatrix.squash_key(key)
        if len(k) > 2:
            raise KeyError(
                "Key formatted incorrectly, must be tuple of <= 2 integers "
                "See PUSOMatrix instead.")
        return k

    @property
    def h(self):
        """h.

        Return a plain dictionary representing the QUSO field values. Each key
        is an integer.

        Returns
        -------
        h : dict.
            Plain dictionary representing the QUSO field values.

        """
        return {k[0]: v for k, v in self.items() if len(k) == 1}

    @property
    def J(self):
        """J.

        Return a plain dictionary representing the QUSO coupling values. Each
        key is an integer.

        Returns
        -------
        J : dict.
            Plain dictionary representing the QUSO coupling values.

        """
        return {k: v for k, v in self.items() if len(k) == 2}



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_pubo_simulation.py.

This file contains the ``PUBOSimulation`` object, which deals with using a
Metropolis algorithm to simulate PUBOs.

"""

from qubovert.utils import pubo_to_puso, spin_to_boolean, boolean_to_spin
from . import PUSOSimulation


__all__ = 'PUBOSimulation',


class PUBOSimulation(PUSOSimulation):
    """PUBOSimulation.

    ``PUBOSimulation`` uses a Metropolis algorithm to simulate a PUBO. The
    PUBO can be "updated" at a temperature ``T``. Thus,
    we can get an idea of the time evolution of a PUBO by creating a
    temperature schedule and seeing how the system evolves. Please note that
    the ``qv.sim.QUBOSimulation`` performs the simulation much faster than the
    ``qv.sim.PUBOSimulation`` object, but of course only works with degree 1 or
    2 models. If your PUBO is degree two (thus a QUBO), then you should use
    the ``QUBOSimulation`` object.

    ``PUBOSimulation`` inherits from ``PUSOSimulation``. In fact,
    ``PUBOSimulation`` just deals internally with converting to and from
    a spin system; all the simulation is done with ``PUSOSimulation``. See
    ``help(qubovert.sim.PUSOSimulation)`` for more details.

    Examples
    --------
    Consider the following example where we minimize an objective function.

    >>> import qubovert as qv
    >>>
    >>> # create the objective function.
    >>> x = [qv.boolean_var(i) for i in range(10)]
    >>> model = sum(x) ** 2
    >>> model.add_constraint_le_zero(x[0] + x[2] - 3 * x[5] - 1, lam=3)
    >>>
    >>> # initial state is all variables equal to 1
    >>> initial_state = {i: 11 for i in range(length)}
    >>> sim = qv.sim.PUBOSimulation(model, initial_state)
    >>>
    >>> # define a schedule. here we simulate at temperature 4 for 25 time
    >>> # steps, then temperature 2 for 25 time steps, then temperature 1 for
    >>> # 10 time steps.
    >>> schedule = (4, 25), (2, 25), (1, 10)
    >>> sim.schedule_update(schedule)
    >>>
    >>> print("final state", sim.state)

    See Also
    --------
    ``qv.sim.PUSOSimulation``, ``qv.sim.QUSOSimulation``,
    ``qv.sim.QUBOSimulation``.

    """

    def __init__(self, P, initial_state=None, memory=0):
        """__init__.

        Parameters
        ----------
        P : dict or type in ``qubovert.BOOLEAN_MODELS``.
            The PUBO to simulate. This should map tuples of boolean variable
            labels to their respective coefficient in the objective function.
            For more information, see the docstrings for any of the models in
            ``qubovert.BOOLEAN_MODELS``.
        initial_state : dict (optional, defaults to None).
            The initial state to start the simulation in. ``initial_state``
            should map boolean label names to their initial values, where each
            value is either 0 or 1. If ``initial_state`` is None, then it
            will be initialized to all 0s.
        memory : int >= 0 (optional, defaults to 0).
            During the simulation, we keep a list of the most recent ``memory``
            states that the simulation was in. These can be accessed with
            ``self.get_past_states(number_of_states)``.

        """
        model = pubo_to_puso(P)
        if initial_state is None:
            initial_state = {v: 0 for v in model._variables}
        super().__init__(model, initial_state, memory)

    @property
    def state(self):
        """state.

        A copy of the current state of the system.

        Returns
        -------
        state : dict.
            Dictionary that maps boolean labels to their values in {0, 1}.

        """
        return spin_to_boolean(self._state)

    def set_state(self, state):
        """set_state.

        Set the state of the spin system to ``state``.

        Parameters
        ----------
        state : dict or iterable.
            ``state`` maps the spin variable labels to their corresponding
            values. In other words ``state[v]`` is the value of variable ``v``.
            A value must be either 0 or 1.

        """
        # if we call super then we get the wrong errors
        state = {v: state[v] for v in self._variables}
        if any(v not in {0, 1} for v in state.values()):
            raise ValueError("State must contain only 0's and 1's")
        self._state = boolean_to_spin(state)



# distutils: language=c
# cython: language_level=3
#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from libc.stdlib cimport malloc, free


cdef extern from "simulate_quso.h":
    void simulate_quso(
        int len_state, int *state, double *h,
        int *num_neighbors, int *neighbors, double *J,
        int len_Ts, double *Ts, int *num_updates,
        int seed
    ) nogil


def c_simulate_quso(state, h, num_neighbors,
                    neighbors, J, schedule, seed):
    """
    Simulate a QUSO with the C source.

    Parameters
    ----------
    state: list of ints.
        `state[i]` is the value of the ith spin, either 1 or -1.
    h : list of floats.
        `h[i]` is the field value on spin `i`.
    num_neighbors : list of ints. 
        `num_neighbors[i]` is the number of neighbors that spin i has.
    neighbors : list of ints.
        ``neighbors[i]`` is the jth neighbor of spin ``k``, where 
        ``j = i - num_neighbors[k-1] - num_neighbors[k-2] - ...``
    J : list of doubles.
        ``J[i]`` is the coupling value between spin ``k`` and 
        ``neighbors[i]``.
    schedule : iterable of tuples.
        Each tuple is a ``T, n`` pairs, where ``n`` is the number of time
        steps to update the simulation at temperature ``T``.
    seed : int. 
        seeds the random number generator If `seed` is a negative integer,
        then we seed the random number generator with `time(NULL)`.
        Otherwise, we use `seed`.

    Returns
    -------
    new_state : list of ints.

    Example
    -------
    `neighbors` and `J` are basically flattened arrays.
    In other words, we flatten the arrays `temp_neighbors` and
    `temp_J`, where `temp_neighbors` points to an array where `temp_neighbors[i][j]`
    is the jth neighbor of spin i, for j=0,...,num_neighbors[i]-1, and similarly,
    `temp_J` points to an array where `temp_J[i][j]` is the coupling value between
    spin i and spin `neighbors[i][j]`, for j=0,...,num_neighbors[i]-1.

    A spin model such as
        -z_0 z_1 + 2*z_1*z_2 + z_0
    must be represented as
        `h = {1., 0, 0}`
        `num_neighbors = {1, 2, 1}`
        `temp_neighbors = {{1}, {0, 2}, {1}}`
        `temp_J = {{-1.},
              {-1, 2},
              {2}}`
        `neighbors = {1, 0, 2, 1}`
        `J = {-1.,
              -1, 2,
              2}`
    """
    # convert all Python types to C
    cdef int c_len_state = len(state)
    cdef int *c_state = <int*>malloc(len(state) * sizeof(int))
    cdef double *c_h = <double*>malloc(len(state) * sizeof(double))
    cdef int *c_num_neighbors = <int*>malloc(len(state) * sizeof(int))
    cdef int *c_neighbors = <int*>malloc(len(neighbors) * sizeof(int))
    cdef double *c_J = <double*>malloc(len(J) * sizeof(double))
    cdef int c_len_Ts = len(schedule)
    cdef double *c_Ts = <double*>malloc(len(schedule) * sizeof(double))
    cdef int *c_num_updates = <int*>malloc(len(schedule) * sizeof(int))
    cdef int c_seed = seed

    for i in range(len(state)):
        c_state[i] = state[i]
        c_h[i] = h[i]
        c_num_neighbors[i] = num_neighbors[i]
    for i in range(len(J)):
        c_neighbors[i] = neighbors[i]
        c_J[i] = J[i]
    for i, (T, n) in enumerate(schedule):
        c_Ts[i] = T
        c_num_updates[i] = n
        if n < 0:
            raise ValueError("Cannot update a negative number of times")

    with nogil:
        simulate_quso(
            c_len_state, c_state, c_h,
            c_num_neighbors, c_neighbors, c_J,
            c_len_Ts, c_Ts, c_num_updates,
            c_seed
        )

    final_state = [c_state[i] for i in range(len(state))]
    free(c_state)
    free(c_h)
    free(c_num_neighbors)
    free(c_neighbors)
    free(c_J)
    free(c_Ts)
    free(c_num_updates)
    return final_state



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""setup.py.

Set up details for ``pip install qubovert`` or ``pip install -e .`` if
installing by source.

"""

import setuptools
from setuptools.command.build_ext import build_ext

try:
    from Cython.Build import cythonize
    USE_CYTHON, ext = True, '.pyx'
except ImportError:
    def cythonize(e): return e
    USE_CYTHON, ext = False, '.c'


with open('README.rst') as f:
    README = f.read()

with open("requirements.txt") as f:
    REQUIREMENTS = [line.strip() for line in f if line.strip()]

# get __version__, __author__, etc.
with open("qubovert/_version.py") as f:
    exec(f.read())


# create the extension for the C file in qubovert.sim.src
extensions = cythonize([
    setuptools.Extension(
        name='qubovert.sim._simulate_quso',
        sources=['./qubovert/sim/_simulate_quso' + ext,
                 './qubovert/sim/src/simulate_quso.c',
                 './qubovert/sim/src/pcg_basic.c'],
        include_dirs=['./qubovert/sim/src/'],
        language='c'
    )
])

setuptools.setup(
    name="qubovert",
    version=__version__,
    author=__author__,
    author_email=__authoremail__,
    description=__description__,
    long_description=README,
    long_description_content_type='text/x-rst',
    url=__sourceurl__,
    license=__license__,
    packages=setuptools.find_packages(exclude=("tests", "docs")),
    ext_modules=extensions,
    test_suite="tests",
    install_requires=REQUIREMENTS,
    zip_safe=False,
    cmdclass=dict(build_ext=build_ext),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Source": __sourceurl__,
        "Docs": __docsurl__
    }
)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_anneal.py.

This file uses the simulation functionality to implement a simulated annealing
algorithm for both boolean and spin models.

"""

from qubovert.utils import (
    puso_value, pubo_to_puso, qubo_to_quso, QUBOVertWarning, boolean_to_spin
)
from . import PUSOSimulation, QUSOSimulation, AnnealResults
import random
import numpy as np

__all__ = (
    'anneal_qubo', 'anneal_quso', 'anneal_pubo', 'anneal_puso',
    'anneal_temperature_range'
)


# anneal temperature range function

def anneal_temperature_range(model, start_flip_prob=0.5,
                             end_flip_prob=0.01, spin=False):
    """anneal_temperature_range.

    Calculate the temperature to start and end an anneal of ``model``, such
    that at the start of the anneal there is a ``start_flip_prob`` probability
    that a bit is flipped despite it being energetically unfavorable, and at
    the end of the anneal there is a ``end_flip_prob`` probability that a bit
    is flipped despite it being energetically unfavorable.

    Parameters
    ----------
    model : dict, or any type in ``qubovert.SPIN_MODELS`` or ``BOOLEAN_MODELS``
        Dictionary mapping tuples of binary labels to their values. See any of
        the docstrings of a type in ``qubovert.SPIN_MODELS`` or
        ``BOOLEAN_MODELS`` for more info.
    start_flip_prob : float in [0, 1) (optional, defaults to 0.5).
        The desired probability that a bit flips despite it being energetically
        unfavorable at the start of the anneal. ``start_flip_prob`` must be
        greater than ``end_flip_prob``.
    end_flip_prob : float in [0, 1) (optional, defaults to 0.01).
        The desired probability that a bit flips despite it being energetically
        unfavorable at the end of the anneal. ``end_flip_prob`` must be
        less than ``start_flip_prob``.
    spin : bool (optional, default to False).
        ``spin`` should be True if ``model`` is a spin model (ie
        ``isinstance(model, qubovert.SPIN_MODELS)``) and should be False if
        ``model`` is a boolean model (ie
        ``isinstance(model, qubovert.BOOLEAN_MODELS)``).

    Returns
    -------
    temp_range : tuple (hot, cold).
        The ``hot`` temperature is the temperature to start the anneal at, and
        the ``cold`` temperature is the temperature to end the anneal at.
        Note that ``hot >= cold``.

    """
    # slight modification of _default_ising_beta_range in
    # https://github.com/dwavesystems/dwave-neal/blob/master/neal/sampler.py

    # raise exception if invalid probabilities
    if any((
        start_flip_prob < 0, start_flip_prob >= 1,
        end_flip_prob < 0, end_flip_prob >= 1,
    )):
        raise ValueError("Flip probabilities must be in [0, 1)")
    elif end_flip_prob > start_flip_prob:
        raise ValueError("The starting flip probability must be greater than "
                         "the ending flip probability.")

    if not spin:
        model = pubo_to_puso(model)

    # if D is a Matrix object or QUBO, PUBO, etc, then variables are defined
    try:
        # don't waste time copying (model.variables), since we never mutate it.
        variables = model._variables
    except AttributeError:
        variables = set(v for k in model for v in k)

    # if the model is empty or just an offset
    if not variables:
        return 0, 0

    factor = 2  # should be this
    # factor = 1  # D-Wave neal does this.

    # calculate the approximate minimum possible change in energy by flipping
    # a single bit.
    min_del_energy = factor * min(abs(c) for k, c in model.items() if k)
    # calculate the approximate maximum possible change in energy by flipping
    # a single bit.
    max_del_energy = factor * max(
        sum(abs(c) for k, c in model.items() if v in k)
        for v in variables
    )

    # now ensure that the bolzmann weight satisfy the desired probabilities.
    # ie exp(-del_energy / T) = prob
    T0 = -max_del_energy / np.log(start_flip_prob) if start_flip_prob else 0
    Tf = -min_del_energy / np.log(end_flip_prob) if end_flip_prob else 0
    return float(T0), float(Tf)


# main spin annealing function

def _anneal_spin(model, spin_simulation, num_anneals=1,
                 anneal_duration=1000, initial_state=None,
                 temperature_range=None, schedule='geometric', seed=None):
    """_anneal_spin.

    Run a simulated annealing algorithm to try to find the minimum of the spin
    model given by ``model``. ``_anneal_spin`` uses a cooling schedule with the
    ``spin_simulation`` object. Please see all of the parameters for details.

    Both ``qv.sim.anneal_puso`` and ``qv.sim.anneal_quso`` run through this
    function. Since ``qv.sim.QUSOSimulation`` is faster than
    ``qv.sim.PUSOSimulation``, we send in different simulation objects for
    ``anneal_quso`` and ``anneal_puso``.

    Parameters
    ----------
    model : dict, or any type in ``qubovert.SPIN_MODELS``.
        Maps spin labels to their values in the Hamiltonian.
        Please see the docstrings of any of the objects in
        ``qubovert.SPIN_MODELS`` to see how ``H`` should be formatted.
    spin_simulation : qv.sim.PUSOSimulation or qv.sim.QUSOSimulation object.
        Should be a ``qv.sim.QUSOSimulation`` object if this function is called
        from ``qv.sim.anneal_quso``, or a ``qv.sim.PUSOSimulation`` object if
        this function is called from ``qv.sim.anneal_puso``.
    num_anneals : int >= 1 (optional, defaults to 1).
        The number of times to run the simulated annealing algorithm.
    anneal_duration : int >= 1 (optional, defaults to 1000).
        The total number of updates to the simulation during the anneal.
        This is related to the amount of time we spend in the cooling schedule.
        If an explicit schedule is provided, then ``anneal_duration`` will be
        ignored.
    initial_state : dict (optional, defaults to None).
        The initial state to start the anneal in. ``initial_state`` must map
        the spin label names to their values in {1, -1}. If ``initial_state``
        is None, then a random state will be chosen to start each anneal.
        Otherwise, ``initial_state`` will be the starting state for all of the
        anneals.
    temperature_range : tuple (optional, defaults to None).
        The temperature to start and end the anneal at.
        ``temperature = (T0, Tf)``. ``T0`` must be >= ``Tf``. To see more
        details on picking a temperature range, please see the function
        ``qubovert.sim.anneal_temperature_range``. If ``temperature_range`` is
        None, then it will by default be set to
        ``T0, Tf = qubovert.sim.anneal_temperature_range(H, spin=True)``.
        Note that a temperature can only be zero if ``schedule`` is explicitly
        given or if ``schedule`` is linear.
    schedule : str or iterable of tuple (optional, defaults to ``'geometric'``)
        What type of cooling schedule to use. If ``schedule == 'linear'``, then
        the cooling schedule will be a linear interpolation between the values
        in ``temperature_range``. If ``schedule == 'geometric'``, then the
        cooling schedule will be a geometric interpolation between the values
        in ``temperature_range``. Otherwise, you can supply an explicit
        schedule. In this case, ``schedule`` should be an iterable of tuples,
        where each tuple is a ``(T, n)`` pair, where ``T`` denotes the
        temperature to update the simulation, and ``n`` denote the number of
        times to update the simulation at that temperature. This schedule
        will be sent directly into the
        ``qubovert.sim.PUSOSimulation.schedule_update`` method.
    seed : number (optional, defaults to None).
        The number to seed Python's builtin ``random`` module with. If
        ``seed is None``, then ``random.seed`` will not be called.

    Returns
    -------
    res : qubovert.sim.AnnealResults object.
        ``res`` contains information on the final states of the simulations.
        See Examples below for an example of how to read from ``res``.
        See ``help(qubovert.sim.AnnealResults)`` for more info.

    Raises
    ------
    ValueError
        If the ``schedule`` argument provided is formatted incorrectly. See the
        Parameters section.
    ValueError
        If the initial temperature is less than the final temperature.

    Warns
    -----
    qubovert.utils.QUBOVertWarning
        If both the ``temperature_range`` and explicit ``schedule`` arguments
        are provided.

    """
    if seed is not None:
        random.seed(seed)

    if schedule in ('linear', 'geometric'):
        T0, Tf = temperature_range or anneal_temperature_range(model,
                                                               spin=True)
        if T0 < Tf:
            raise ValueError("The final temperature must be less than the "
                             "initial temperature")

        # in the case that H is empty or just an offset and the user didn't
        # supply a temperature range, then T0 and Tf will be 0.
        if temperature_range is None and T0 == Tf == 0:
            T0 = Tf = 1
        Ts = (
            np.linspace(T0, Tf, anneal_duration) if schedule == 'linear' else
            np.geomspace(T0, Tf, anneal_duration)
        )
        schedule = tuple((T, 1) for T in Ts)
    elif isinstance(schedule, str):
        raise ValueError(
            "Invalid schedule. Must be either 'linear', 'geometric', or an "
            "explicit temperature schedule. See the docstring for more info."
        )
    elif temperature_range:
        QUBOVertWarning.warn(
            "Both a temperature range and an explicit schedule was provided. "
            "The temperature range will be ignored and the schedule used "
            "instead."
        )

    sim = spin_simulation(model, initial_state)

    result = AnnealResults(True)
    for _ in range(num_anneals):
        if initial_state is None:
            sim.set_state({v: random.choice((-1, 1)) for v in sim._variables})
        sim.schedule_update(
            schedule,
            seed=random.randint(0, 1 << 16) if seed is not None else None
        )
        state = sim.state
        result.add_state(state, puso_value(state, model))
        sim.reset()

    return result


# annealing functions

def anneal_puso(H, num_anneals=1, anneal_duration=1000, initial_state=None,
                temperature_range=None, schedule='geometric', seed=None):
    """anneal_puso.

    Run a simulated annealing algorithm to try to find the minimum of the PUSO
    given by ``H``. ``anneal_puso`` uses a cooling schedule with the
    ``qubovert.sim.PUSOSimulation`` object. Please see all of the parameters
    for details.

    **Please note** that the ``qv.sim.anneal_quso`` function performs much
    faster than the ``qv.sim.anneal_puso`` function since the former is written
    in C and wrapped in Python. If your system has degree 2 or less, then you
    should use the ``qv.sim.anneal_quso`` function!

    Parameters
    ----------
    H : dict, or any type in ``qubovert.SPIN_MODELS``.
        Maps spin labels to their values in the Hamiltonian.
        Please see the docstrings of any of the objects in
        ``qubovert.SPIN_MODELS`` to see how ``H`` should be formatted.
    num_anneals : int >= 1 (optional, defaults to 1).
        The number of times to run the simulated annealing algorithm.
    anneal_duration : int >= 1 (optional, defaults to 1000).
        The total number of updates to the simulation during the anneal.
        This is related to the amount of time we spend in the cooling schedule.
        If an explicit schedule is provided, then ``anneal_duration`` will be
        ignored.
    initial_state : dict (optional, defaults to None).
        The initial state to start the anneal in. ``initial_state`` must map
        the spin label names to their values in {1, -1}. If ``initial_state``
        is None, then a random state will be chosen to start each anneal.
        Otherwise, ``initial_state`` will be the starting state for all of the
        anneals.
    temperature_range : tuple (optional, defaults to None).
        The temperature to start and end the anneal at.
        ``temperature = (T0, Tf)``. ``T0`` must be >= ``Tf``. To see more
        details on picking a temperature range, please see the function
        ``qubovert.sim.anneal_temperature_range``. If ``temperature_range`` is
        None, then it will by default be set to
        ``T0, Tf = qubovert.sim.anneal_temperature_range(H, spin=True)``.
        Note that a temperature can only be zero if ``schedule`` is explicitly
        given or if ``schedule`` is linear.
    schedule : str or iterable of tuple (optional, defaults to ``'geometric'``)
        What type of cooling schedule to use. If ``schedule == 'linear'``, then
        the cooling schedule will be a linear interpolation between the values
        in ``temperature_range``. If ``schedule == 'geometric'``, then the
        cooling schedule will be a geometric interpolation between the values
        in ``temperature_range``. Otherwise, you can supply an explicit
        schedule. In this case, ``schedule`` should be an iterable of tuples,
        where each tuple is a ``(T, n)`` pair, where ``T`` denotes the
        temperature to update the simulation, and ``n`` denote the number of
        times to update the simulation at that temperature. This schedule
        will be sent directly into the
        ``qubovert.sim.PUSOSimulation.schedule_update`` method.
    seed : number (optional, defaults to None).
        The number to seed Python's builtin ``random`` module with. If
        ``seed is None``, then ``random.seed`` will not be called.

    Returns
    -------
    res : qubovert.sim.AnnealResults object.
        ``res`` contains information on the final states of the simulations.
        See Examples below for an example of how to read from ``res``.
        See ``help(qubovert.sim.AnnealResults)`` for more info.

    Raises
    ------
    ValueError
        If the ``schedule`` argument provided is formatted incorrectly. See the
        Parameters section.
    ValueError
        If the initial temperature is less than the final temperature.

    Warns
    -----
    qubovert.utils.QUBOVertWarning
        If both the ``temperature_range`` and explicit ``schedule`` arguments
        are provided.

    Example
    -------
    Consider the example of finding the ground state of the 1D
    antiferromagnetic Ising chain of length 5.

    >>> import qubovert as qv
    >>>
    >>> H = sum(qv.spin_var(i) * qv.spin_var(i+1) for i in range(4))
    >>> anneal_res = qv.sim.anneal_puso(H, num_anneals=3)
    >>>
    >>> print(anneal_res.best.value)
    -4
    >>> print(anneal_res.best.state)
    {0: 1, 1: -1, 2: 1, 3: -1, 4: 1}
    >>> # now sort the results
    >>> anneal_res.sort_by_value()
    >>>
    >>> # now iterate through all of the results in the sorted order
    >>> for res in anneal_res:
    >>>     print(res.value, res.state)
    -4, {0: 1, 1: -1, 2: 1, 3: -1, 4: 1}
    -4, {0: -1, 1: 1, 2: -1, 3: 1, 4: -1}
    -4, {0: 1, 1: -1, 2: 1, 3: -1, 4: 1}

    """
    return _anneal_spin(
        H, PUSOSimulation, num_anneals, anneal_duration, initial_state,
        temperature_range, schedule, seed
    )


def anneal_pubo(P, num_anneals=1, anneal_duration=1000, initial_state=None,
                temperature_range=None, schedule='geometric', seed=None):
    """anneal_pubo.

    Run a simulated annealing algorithm to try to find the minimum of the PUBO
    given by ``P``. ``anneal_pubo`` uses a cooling schedule with the
    ``qubovert.sim.PUBOSimulation`` object. Please see all of the parameters
    for details.

    **Please note** that the ``qv.sim.anneal_qubo`` function performs much
    faster than the ``qv.sim.anneal_pubo`` function since the former is written
    in C and wrapped in Python. If your system has degree 2 or less, then you
    should use the ``qv.sim.anneal_qubo`` function!

    Parameters
    ----------
    P : dict, or any type in ``qubovert.BOOLEAN_MODELS``.
        Maps boolean labels to their values in the objective function.
        Please see the docstrings of any of the objects in
        ``qubovert.BOOLEAN_MODELS`` to see how ``P`` should be formatted.
    num_anneals : int >= 1 (optional, defaults to 1).
        The number of times to run the simulated annealing algorithm.
    anneal_duration : int >= 1 (optional, defaults to 1000).
        The total number of updates to the simulation during the anneal.
        This is related to the amount of time we spend in the cooling schedule.
        If an explicit schedule is provided, then ``anneal_duration`` will be
        ignored.
    initial_state : dict (optional, defaults to None).
        The initial state to start the anneal in. ``initial_state`` must map
        the boolean label names to their values in {0, 1}. If ``initial_state``
        is None, then a random state will be chosen to start each anneal.
        Otherwise, ``initial_state`` will be the starting state for all of the
        anneals.
    temperature_range : tuple (optional, defaults to None).
        The temperature to start and end the anneal at.
        ``temperature = (T0, Tf)``. ``T0`` must be >= ``Tf``. To see more
        details on picking a temperature range, please see the function
        ``qubovert.sim.anneal_temperature_range``. If ``temperature_range`` is
        None, then it will by default be set to
        ``T0, Tf = qubovert.sim.anneal_temperature_range(P, spin=False)``.
    schedule : str or iterable of tuple (optional, defaults to ``'geometric'``)
        What type of cooling schedule to use. If ``schedule == 'linear'``, then
        the cooling schedule will be a linear interpolation between the values
        in ``temperature_range``. If ``schedule == 'geometric'``, then the
        cooling schedule will be a geometric interpolation between the values
        in ``temperature_range``. Otherwise, you can supply an explicit
        schedule. In this case, ``schedule`` should be an iterable of tuples,
        where each tuple is a ``(T, n)`` pair, where ``T`` denotes the
        temperature to update the simulation, and ``n`` denote the number of
        times to update the simulation at that temperature. This schedule
        will be sent directly into the
        ``qubovert.sim.PUBOSimulation.schedule_update`` method.
    seed : number (optional, defaults to None).
        The number to seed Python's builtin ``random`` module with. If
        ``seed is None``, then ``random.seed`` will not be called.

    Returns
    -------
    res : qubovert.sim.AnnealResults object.
        ``res`` contains information on the final states of the simulations.
        See Examples below for an example of how to read from ``res``.
        See ``help(qubovert.sim.AnnealResults)`` for more info.

    Raises
    ------
    ValueError
        If the ``schedule`` argument provided is formatted incorrectly. See the
        Parameters section.
    ValueError
        If the initial temperature is less than the final temperature.

    Warns
    -----
    qubovert.utils.QUBOVertWarning
        If both the ``temperature_range`` and explicit ``schedule`` arguments
        are provided.

    Example
    -------
    Consider the example of finding the ground state of the 1D
    antiferromagnetic Ising chain of length 5 in boolean form.

    >>> import qubovert as qv
    >>>
    >>> H = sum(qv.spin_var(i) * qv.spin_var(i+1) for i in range(4))
    >>> P = H.to_pubo()
    >>> anneal_res = qv.sim.anneal_pubo(P, num_anneals=3)
    >>>
    >>> print(anneal_res.best.value)
    -4
    >>> print(anneal_res.best.state)
    {0: 0, 1: 1, 2: 0, 3: 1, 4: 0}
    >>> # now sort the results
    >>> anneal_res.sort_by_value()
    >>>
    >>> # now iterate through all of the results in the sorted order
    >>> for res in anneal_res:
    >>>     print(res.value, res.state)
    -4, {0: 0, 1: 1, 2: 0, 3: 1, 4: 0}
    -4, {0: 1, 1: 0, 2: 1, 3: 0, 4: 1}
    -4, {0: 0, 1: 1, 2: 0, 3: 1, 4: 0}

    """
    return anneal_puso(
        pubo_to_puso(P), num_anneals, anneal_duration,
        None if initial_state is None else boolean_to_spin(initial_state),
        temperature_range, schedule, seed
    ).to_boolean()


def anneal_quso(L, num_anneals=1, anneal_duration=1000, initial_state=None,
                temperature_range=None, schedule='geometric', seed=None):
    """anneal_quso.

    Run a simulated annealing algorithm to try to find the minimum of the QUSO
    given by ``L``. ``anneal_quso`` uses a cooling schedule with the
    ``qubovert.sim.PUSOSimulation`` object. Please see all of the parameters
    for details.

    Parameters
    ----------
    L : dict, ``qubovert.utils.QUSOMatrix`` or ``qubovert.QUSO``.
        Maps spin labels to their values in the objective function.
        Please see the docstring of ``qubovert.QUSO`` for more info on how to
        format ``L``.
    num_anneals : int >= 1 (optional, defaults to 1).
        The number of times to run the simulated annealing algorithm.
    anneal_duration : int >= 1 (optional, defaults to 1000).
        The total number of updates to the simulation during the anneal.
        This is related to the amount of time we spend in the cooling schedule.
        If an explicit schedule is provided, then ``anneal_duration`` will be
        ignored.
    initial_state : dict (optional, defaults to None).
        The initial state to start the anneal in. ``initial_state`` must map
        the spin label names to their values in {1, -1}. If ``initial_state``
        is None, then a random state will be chosen to start each anneal.
        Otherwise, ``initial_state`` will be the starting state for all of the
        anneals.
    temperature_range : tuple (optional, defaults to None).
        The temperature to start and end the anneal at.
        ``temperature = (T0, Tf)``. ``T0`` must be >= ``Tf``. To see more
        details on picking a temperature range, please see the function
        ``qubovert.sim.anneal_temperature_range``. If ``temperature_range`` is
        None, then it will by default be set to
        ``T0, Tf = qubovert.sim.anneal_temperature_range(L, spin=True)``.
    schedule : str or iterable of tuple (optional, defaults to ``'geometric'``)
        What type of cooling schedule to use. If ``schedule == 'linear'``, then
        the cooling schedule will be a linear interpolation between the values
        in ``temperature_range``. If ``schedule == 'geometric'``, then the
        cooling schedule will be a geometric interpolation between the values
        in ``temperature_range``. Otherwise, you can supply an explicit
        schedule. In this case, ``schedule`` should be an iterable of tuples,
        where each tuple is a ``(T, n)`` pair, where ``T`` denotes the
        temperature to update the simulation, and ``n`` denote the number of
        times to update the simulation at that temperature. This schedule
        will be sent directly into the
        ``qubovert.sim.PUSOSimulation.schedule_update`` method.
    seed : number (optional, defaults to None).
        The number to seed Python's builtin ``random`` module with. If
        ``seed is None``, then ``random.seed`` will not be called.

    Returns
    -------
    res : qubovert.sim.AnnealResults object.
        ``res`` contains information on the final states of the simulations.
        See Examples below for an example of how to read from ``res``.
        See ``help(qubovert.sim.AnnealResults)`` for more info.

    Raises
    ------
    ValueError
        If the ``schedule`` argument provided is formatted incorrectly. See the
        Parameters section.
    ValueError
        If the initial temperature is less than the final temperature.

    Warns
    -----
    qubovert.utils.QUBOVertWarning
        If both the ``temperature_range`` and explicit ``schedule`` arguments
        are provided.

    Example
    -------
    Consider the example of finding the ground state of the 1D
    antiferromagnetic Ising chain of length 5.

    >>> import qubovert as qv
    >>>
    >>> H = sum(qv.spin_var(i) * qv.spin_var(i+1) for i in range(4))
    >>> anneal_res = qv.sim.anneal_quso(H, num_anneals=3)
    >>>
    >>> print(anneal_res.best.value)
    -4
    >>> print(anneal_res.best.state)
    {0: 1, 1: -1, 2: 1, 3: -1, 4: 1}
    >>> # now sort the results
    >>> anneal_res.sort_by_value()
    >>>
    >>> # now iterate through all of the results in the sorted order
    >>> for res in anneal_res:
    >>>     print(res.value, res.state)
    -4, {0: 1, 1: -1, 2: 1, 3: -1, 4: 1}
    -4, {0: -1, 1: 1, 2: -1, 3: 1, 4: -1}
    -4, {0: 1, 1: -1, 2: 1, 3: -1, 4: 1}

    """
    return _anneal_spin(
        L, QUSOSimulation, num_anneals, anneal_duration, initial_state,
        temperature_range, schedule, seed
    )


def anneal_qubo(Q, num_anneals=1, anneal_duration=1000, initial_state=None,
                temperature_range=None, schedule='geometric', seed=None):
    """anneal_qubo.

    Run a simulated annealing algorithm to try to find the minimum of the QUBO
    given by ``Q``. ``anneal_qubo`` uses a cooling schedule with the
    ``qubovert.sim.PUBOSimulation`` object. Please see all of the parameters
    for details.

    Parameters
    ----------
    Q : dict, ``qubovert.utils.QUBOMatrix`` or ``qubovert.QUBO``.
        Maps boolean labels to their values in the objective function.
        Please see the docstring of ``qubovert.QUBO`` for more info on how to
        format ``Q``.
    num_anneals : int >= 1 (optional, defaults to 1).
        The number of times to run the simulated annealing algorithm.
    anneal_duration : int >= 1 (optional, defaults to 1000).
        The total number of updates to the simulation during the anneal.
        This is related to the amount of time we spend in the cooling schedule.
        If an explicit schedule is provided, then ``anneal_duration`` will be
        ignored.
    initial_state : dict (optional, defaults to None).
        The initial state to start the anneal in. ``initial_state`` must map
        the boolean label names to their values in {0, 1}. If ``initial_state``
        is None, then a random state will be chosen to start each anneal.
        Otherwise, ``initial_state`` will be the starting state for all of the
        anneals.
    temperature_range : tuple (optional, defaults to None).
        The temperature to start and end the anneal at.
        ``temperature = (T0, Tf)``. ``T0`` must be >= ``Tf``. To see more
        details on picking a temperature range, please see the function
        ``qubovert.sim.anneal_temperature_range``. If ``temperature_range`` is
        None, then it will by default be set to
        ``T0, Tf = qubovert.sim.anneal_temperature_range(Q, spin=False)``.
    schedule : str or iterable of tuple (optional, defaults to ``'geometric'``)
        What type of cooling schedule to use. If ``schedule == 'linear'``, then
        the cooling schedule will be a linear interpolation between the values
        in ``temperature_range``. If ``schedule == 'geometric'``, then the
        cooling schedule will be a geometric interpolation between the values
        in ``temperature_range``. Otherwise, you can supply an explicit
        schedule. In this case, ``schedule`` should be an iterable of tuples,
        where each tuple is a ``(T, n)`` pair, where ``T`` denotes the
        temperature to update the simulation, and ``n`` denote the number of
        times to update the simulation at that temperature. This schedule
        will be sent directly into the
        ``qubovert.sim.PUBOSimulation.schedule_update`` method.
    seed : number (optional, defaults to None).
        The number to seed Python's builtin ``random`` module with. If
        ``seed is None``, then ``random.seed`` will not be called.

    Returns
    -------
    res : qubovert.sim.AnnealResults object.
        ``res`` contains information on the final states of the simulations.
        See Examples below for an example of how to read from ``res``.
        See ``help(qubovert.sim.AnnealResults)`` for more info.

    Raises
    ------
    ValueError
        If the ``schedule`` argument provided is formatted incorrectly. See the
        Parameters section.
    ValueError
        If the initial temperature is less than the final temperature.

    Warns
    -----
    qubovert.utils.QUBOVertWarning
        If both the ``temperature_range`` and explicit ``schedule`` arguments
        are provided.

    Example
    -------
    Consider the example of finding the ground state of the 1D
    antiferromagnetic Ising chain of length 5 in boolean form.

    >>> import qubovert as qv
    >>>
    >>> H = sum(qv.spin_var(i) * qv.spin_var(i+1) for i in range(4))
    >>> Q = H.to_qubo()
    >>> anneal_res = qv.sim.anneal_qubo(Q, num_anneals=3)
    >>>
    >>> print(anneal_res.best.value)
    -4
    >>> print(anneal_res.best.state)
    {0: 0, 1: 1, 2: 0, 3: 1, 4: 0}
    >>> # now sort the results
    >>> anneal_res.sort_by_value()
    >>>
    >>> # now iterate through all of the results in the sorted order
    >>> for res in anneal_res:
    >>>     print(res.value, res.state)
    -4, {0: 0, 1: 1, 2: 0, 3: 1, 4: 0}
    -4, {0: 1, 1: 0, 2: 1, 3: 0, 4: 1}
    -4, {0: 0, 1: 1, 2: 0, 3: 1, 4: 0}

    """
    return anneal_quso(
        qubo_to_quso(Q), num_anneals, anneal_duration,
        None if initial_state is None else boolean_to_spin(initial_state),
        temperature_range, schedule, seed
    ).to_boolean()



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_qubomatrix.py.

This file contains the QUBOMatrix object.

"""

import numpy as np
from . import PUBOMatrix


__all__ = 'QUBOMatrix', 'matrix_to_qubo', 'qubo_to_matrix'


class QUBOMatrix(PUBOMatrix):
    """QUBOMatrix.

    ``QUBOMatrix`` inherits some methods from ``PUBOMatrix``, see
    ``help(qubovert.utils.PUBOMattrix)``.

    A class to handle QUBO matrices. It is the same thing as a dictionary
    with some methods modified. Note that each key must be a tuple of two
    integers >= 0.

    One method is that values will always default to 0. Consider the following
    example:

    >>> d = QUBOMatrix()
    >>> print(d[(0,)]) # will print 0
    >>> d[(0,)] += 1
    >>> print(d) # will print {(0,): 1}

    Compared to an ordinary dictionary.

    >>> g = dict()
    >>> print(g[(0,)]) # will raise KeyError
    >>> g[(0,)] += 1 # will raise KeyError, since (0,) was never set

    One method of QUBOMatrix is that it will always keep the QUBO
    upper triangular! Consider the following example:

    >>> d = QUBOMatrix()
    >>> d[(1, 0)] += 2
    >>> print(d)
    >>> # will print {(0, 1): 2}

    One method is that if we set an item to 0, it will be removed. Consider
    the following example:

    >>> d = QUBOMatrix()
    >>> d[(0,)] += 1
    >>> d[(0,)] -= 1
    >>> print(d) # will print {}

    One method is that if we initialize QUBOMatrix with a previous dictionary
    it will be reinitialized to ensure that the QUBOMatrix is upper
    triangular and contains no zero values. Consider the following example:

    >>> d = QUBOMatrix({(0, 0): 1, (1, 0): 2, (2, 0): 0})
    >>> print(d) # will print {(0, 0): 1, (0, 1): 2}

    We also change the update method so that it follows all the conventions.

    >>> d = QUBOMatrix({(0, 0): 1, (0, 1): 2})
    >>> d.update({(0, 0): 0, (1, 0): 1, (1, 1): -1})
    >>> print(d)  # will print {(0, 1): 1, (1,): -1}

    We also include arithmetic, addition, subtraction, scalar division,
    multiplication, and all those in place. For example,

    >>> d = QUBOMatrix((0, 0)=1, (0, 1)=-2)
    >>> g = d + {(0, 0): -1}
    >>> print(g) # will print {(0, 1): -2}
    >>> g *= 4
    >>> print(g) # will print {(0, 1): -8}
    >>> g -= {(0, 1): -8}
    >>> print(g) # will print {}

    >>> d = QUBOMatrix({(0, 0): 1, (0, 1): -1})
    >>> g = {(0,): -1, (1,): 1}
    >>> d *= g
    >>> print(d)
    {(0,): -1, (0, 1): 1}

    >>> d = QUBOMatrix({(0, 0): 1, (0, 1): -1})
    >>> print(d ** 2 == d * d)
    True

    Adding or subtracting constants will update the () element of the
    dict.

    >>> d = QUBOMatrix()
    >>> d += 5
    >>> print(d)
    {(): 5}

    Finally, if you try to access a key out of order, it will sort and squash
    the key. Be careful with this, it can cause unexpected behavior if you
    don't know it. For example,

    >>> d = QUBOMatrix()
    >>> d[(0, 1)] += 2
    >>> print(d[(1, 0)])  # will print 2

    >>> d = QUBOMatrix()
    >>> d[(0, 0)] += 2
    >>> print(d[(0,)])  # will print 2
    >>> print(d[(0, 0)])  # will print 2
    >>> print(d)  # will print {(0,): 2}

    """

    @staticmethod
    def _check_key_valid(key):
        """_check_key_valid.

        Internal method to check if an input key to the dictionary is valid.
        Checks to see if ``key`` is a tuple of non negative integers with <=
        2 unique integers.

        Parameters
        ----------
        key : anything, but must be a tuple to be valid.

        Returns
        -------
        k : the squashed key.

        Raises
        ------
        KeyError if the key is invalid.

        """
        k = PUBOMatrix.squash_key(key)
        if len(k) > 2:
            raise KeyError(
                "Key formatted incorrectly, must be tuple of <= 2 integers "
                "See PUBOMatrix instead.")
        return k

    @property
    def Q(self):
        """Q.

        Return a plain dictionary representing the QUBO. Each key is a tuple
        of two integers, ie (1, 1) corresponds to (1,). Note that the offset
        in the QUBOMatrix is ignored (ie the value corresponding to the key
        ()). See the ``offset`` property to access it.

        Returns
        -------
        Q : dict.
            Plain dictionary representing the QUBO in standard form.

        """
        return {k * (3 - len(k)): v for k, v in self.items() if k}


def matrix_to_qubo(matrix):
    r"""matrix_to_qubo.

    Convert a matrix to a QUBO dictionary.

    Parameters
    ----------
    matrix : list of lists or 2-dimensional numpy array.
        ``matrix[i][j]`` is equal to :math:`Q_{ij}`.

    Return
    ------
    Q : qubovert.utils.QUBOMatrix object.
       The upper triangular QUBO dictionary. See
       ``help(qubovert.utils.QUBOMatrix)``.

    """
    if not isinstance(matrix, np.ndarray):
        matrix = np.array(matrix)

    if len(matrix.shape) != 2 or matrix.shape[0] != matrix.shape[1]:
        raise ValueError("Input matrix must be square and two-dimensional")

    Q = QUBOMatrix()
    for i in range(matrix.shape[0]):
        for j in range(matrix.shape[1]):
            Q[(i, j)] += matrix[i][j]

    return Q


def qubo_to_matrix(Q, symmetric=False, array=True):
    r"""qubo_to_matrix.

    Convert a QUBO dictionary to its matrix form. The indices of the ``Q``
    dictionary should be integers from 0 to ``n-1``, where there are ``n``
    binary variables in the QUBO problem.

    Parameters
    ----------
    Q : dict or qubovert.utils.QUBOMatrix object.
        Input QUBO dictionary, where ``Q[(i, j)]`` corresponds to
        :math:`Q_{ij}`.
    symmetric : bool (optional, defaults to False).
        Whether the returned matrix should be symmetric or upper-triangular.
        If ``symmetric`` is True, then the matrix will be symmetric, ie
        ``matrix[i][j] == matrix[j][i]``. Otherwise, it will be
        upper-triangular, ie ``marix[i][j] == 0`` if ``i > j``.
    array : bool (optional, defaults to True).
        Whether the returned matrix should be a numpy array or list of lists.
        If ``array`` is True, then it will be a numpy array, otherwise, it
        will be a list of lists.

    Return
    ------
    matrix : numpy array or list of lists.
        The matrix representing the QUBO. See the arguments ``symmetric`` and
        ``array`` for info on the return type of ``matrix``.

    """
    if not Q:
        raise ValueError("QUBO dictionary is empty")
    elif not isinstance(Q, QUBOMatrix):
        Q = QUBOMatrix(Q)

    if Q[()] != 0:
        raise ValueError("QUBO cannot have a constant when converting "
                         "to a matrix")

    matrix = np.zeros((Q.max_index+1,)*2)
    for k, v in Q.items():
        if len(k) == 1:
            matrix[k[0]][k[0]] = v
        elif symmetric:
            i, j = k
            matrix[i][j] = v / 2
            matrix[j][i] = v / 2
        else:
            i, j = k
            matrix[i][j] = v

    if not array:
        return matrix.tolist()
    return matrix



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_puso_simulation.py.

This file contains the ``PUSOSimulation`` object, which deals with using a
Metropolis algorithm to simulate PUSOs.

"""

from qubovert.utils import puso_value
import random
from numpy import exp


__all__ = 'PUSOSimulation',


class PUSOSimulation:
    """PUSOSimulation.

    ``PUSOSimulation`` uses a Metropolis algorithm to simulate a PUSO.
    The spin system can be "updated" at a temperature ``T``. Thus, we can get
    an idea of the time evolution of a spin system by creating a temperature
    schedule and seeing how the system evolves. Please note that the
    ``qv.sim.QUSOSimulation`` performs the simulation much faster than the
    ``qv.sim.PUSOSimulation`` object, but of course only works with degree 1 or
    2 models. If your PUSO is degree two (thus a QUSO), then you should use
    the ``QUSOSimulation`` object.

    Examples
    --------
    Consider the example of the ferromagnetic chain.

    >>> import qubovert as qv
    >>>
    >>> length = 50
    >>> spin_system = sum(
    >>>     -qv.spin_var(i) * qv.spin_var(i+1) * qv.spin_var(i+2)
    >>>     for i in range(length-2)
    >>> )
    >>>
    >>> # initial state is all spin down
    >>> initial_state = {i: -1 for i in range(length)}
    >>> sim = qv.sim.PUSOSimulation(spin_system, initial_state)
    >>>
    >>> # define a schedule. here we simulate at temperature 4 for 25 time
    >>> # steps, then temperature 2 for 25 time steps, then temperature 1 for
    >>> # 10 time steps.
    >>> schedule = (4, 25), (2, 25), (1, 10)
    >>> sim.schedule_update(schedule)
    >>>
    >>> print("final state", sim.state)
    >>> print("last 30 states", sim.get_past_states(30))

    See Also
    --------
    ``qv.sim.PUBOSimulation``, ``qv.sim.QUSOSimulation``,
    ``qv.sim.QUBOSimulation``.

    """

    def __init__(self, H, initial_state=None, memory=0):
        """__init__.

        Parameters
        ----------
        H : dict or type in ``qubovert.SPIN_MODELS``.
            The PUSO to simulate. This should map tuples of spin variable
            labels to their respective coefficient in the Hamiltonian. For more
            information, see the docstrings for any of the models in
            ``qubovert.SPIN_MODELS``.
        initial_state : dict (optional, defaults to None).
            The initial state to start the simulation in. ``initial_state``
            should map spin label names to their initial values, where each
            value is either 1 or -1. If ``initial_state`` is None, then it
            will be initialized to all 1s.
        memory : int >= 0 (optional, defaults to 0).
            During the simulation, we keep a list of the most recent ``memory``
            states that the simulation was in. These can be accessed with
            ``self.get_past_states(number_of_states)``.

        """
        # if model is a Matrix object or PUSO, etc,
        # then variables will be defined.
        # variables must be a list so it can be used with random.choices in
        # the update method.
        try:
            self._variables = list(H._variables)
        except AttributeError:
            if isinstance(initial_state, dict):
                self._variables = list(initial_state.keys())
            else:
                self._variables = list({v for k in H for v in k})

        self._initial_state = (
            initial_state.copy() if initial_state is not None else
            {v: 1 for v in self._variables}
        )
        self.set_state(self._initial_state)

        # keep track of the most recent states
        self._past_states = []
        # how many previous states to remember
        self._memory = memory

        # create a dictionary mapping each bit to the graph that it affects.
        self._subgraphs = {v: {} for v in self._variables}
        for k, c in H.items():
            for v in k:
                self._subgraphs[v][k] = c

    def __str__(self):
        """__str__.

        Return
        ------
        s : str.

        """
        return self.__class__.__name__ + "(memory=%d)" % self._memory

    @property
    def memory(self):
        """memory.

        Returns
        -------
        memory : int >= 0 (optional, defaults to 0).
            During the simulation, we keep a list of the most recent ``memory``
            states that the simulation was in. These can be accessed with
            ``self.get_past_states(number_of_states)``.

        """
        return self._memory

    @property
    def state(self):
        """state.

        A copy of the current state of the system.

        Returns
        -------
        state : dict.
            Dictionary that maps spin labels to their values in {1, -1}.

        """
        return self._state.copy()

    @property
    def initial_state(self):
        """initial_state.

        A copy of the initial state of the system.

        Returns
        -------
        initial_state : dict.
            Dictionary that maps binary labels to their values.

        """
        return self._initial_state.copy()

    def set_state(self, state):
        """set_state.

        Set the state of the spin system to ``state``.

        Parameters
        ----------
        state : dict or iterable.
            ``state`` maps the spin variable labels to their corresponding
            values. In other words ``state[v]`` is the value of variable ``v``.
            A value must be either 1 or -1.

        """
        self._state = {v: state[v] for v in self._variables}
        if any(v not in {1, -1} for v in self._state.values()):
            raise ValueError("State must contain only 1's and -1's")

    def reset(self):
        """reset.

        Reset the simulation back to its original state.

        """
        self._past_states = []
        self.set_state(self._initial_state)

    def get_past_states(self, num_states=None):
        """get_past_states.

        Return the previous ``num_states`` states of the system (if that many
        exist; ``self`` only stores up the previous ``self.memory`` states).

        Parameters
        ----------
        num_states : int (optional, defaults to None).
            The number of previous update steps to include. If ``num_states``
            is None, then all the previous states in memory will be returned.

        Returns
        -------
        states : list of dicts.
            Each dict maps binary labels to their values.

        """
        if num_states == 1:
            return [self.state]
        elif num_states is None:
            num_states = self._memory
        return [
            s.copy() for s in self._past_states[-num_states+1:]
        ] + [self.state]

    def _add_past_state(self):
        """_add_past_state.

        Add the current state to the ``past_states`` memory. If there is no
        more memory left (see ``self.memory``) then remove the oldest state.

        Parameters
        ----------
        state : dict.
            Maps binary labels to their values.

        """
        if self._memory:
            self._past_states.append(self.state)
            if len(self._past_states) > self._memory:
                self._past_states.pop(0)

    def _flip_bit(self, label):
        """_flip_bit.

        Flip the spin labeled by ``label`` in the internal state.

        Parameters
        ----------
        label : hashable object.
            The label of the spin to flip.

        """
        self._state[label] *= -1

    def schedule_update(self, schedule, seed=None):
        """schedule_update.

        Update the simulation with a schedule.

        Parameters
        ----------
        schedule : iterable of tuples.
            Each element in ``schedule`` is a pair ``(T, n)`` which designates
            a temperature and a number of updates. See `Notes` below.
        seed : number (optional, defaults to None).
            The number to seed ``random`` with. If ``seed is None``, then
            ``random.seed`` will not be called.

        Notes
        -----
        The following two code blocks perform exactly the same thing.

        >>> sim = PUSOSimulation(10)
        >>> for T in (3, 2):
        >>>     sim.update(T, 100)
        >>> sim.update(1, 50)

        >>> sim = PUSOSimulation(10)
        >>> schedule = (3, 100), (2, 100), (1, 50)
        >>> sim.schedule_update(schedule)

        """
        if seed is not None:
            random.seed(seed)
        for T, n in schedule:
            self.update(T, n)

    def update(self, T, num_updates=1, seed=None):
        """update.

        Update the simulation at temperature ``T``. Updates the internal state.

        Parameters
        ----------
        T : number >= 0.
            Temperature.
        num_updates : int >= 1 (optional, defaults to 1).
            The number of times to update the simulation at the temperature.
        seed : number (optional, defaults to None).
            The number to seed ``random`` with. If ``seed is None``, then
            ``random.seed`` will not be called.

        """
        if seed is not None:
            random.seed(seed)

        if num_updates < 0:
            raise ValueError("Cannot update a negative number of times")
        elif num_updates > 1:
            for _ in range(num_updates):
                self.update(T)
        elif num_updates == 1:
            self._add_past_state()

            # for i in self._variables:  # dwave does this, why?

            # sample without replacement
            # for i in random.sample(self._variables, len(self._variables)):

            # sample with replacement. this is more valid because it results in
            # a totally memoryless simulation
            for i in random.choices(self._variables, k=len(self._variables)):
                # the change in energy from flipping variable i is equal to
                # -2 * (the energy of the subgraph depending on i)
                dE = -2 * puso_value(self._state, self._subgraphs[i])
                if dE <= 0 or (T and random.random() < exp(-dE / T)):
                    self._flip_bit(i)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``cycles`` contains many NP cycles problems.

Here we have the QUBO/QUSO conversions for common cycle probems,
The conversions are based on [Lucas].

References
----------
.. [Lucas] Andrew Lucas. Ising formulations of many np problems. Frontiers in
   Physics, 2:5, 2014.

"""

__all__ = ()

name = "cycles"
